from datetime import datetime
from itertools import chain
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

from pytezos.context.abstract import AbstractContext
from pytezos.context.abstract import get_originated_address
from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.key import Key
from pytezos.logging import logger
from pytezos.michelson.forge import forge_micheline
from pytezos.michelson.forge import forge_script_expr
from pytezos.michelson.micheline import get_script_section
from pytezos.michelson.micheline import get_script_sections
from pytezos.operation import DEFAULT_OPERATIONS_TTL
from pytezos.operation import MAX_OPERATIONS_TTL
from pytezos.rpc.errors import RpcError
from pytezos.rpc.shell import ShellQuery

DEFAULT_IPFS_GATEWAY = 'https://ipfs.io/ipfs'


class ExecutionContext(AbstractContext):
    def __init__(
        self,
        amount=None,
        chain_id=None,
        protocol=None,
        source=None,
        sender=None,
        balance=None,
        block_id=None,
        now=None,
        level=None,
        voting_power=None,
        total_voting_power=None,
        min_block_time=None,
        key=None,
        shell=None,
        address=None,
        counter=None,
        script=None,
        tzt=False,
        mode=None,
        ipfs_gateway=None,
        global_constants=None,
        view_results=None,
    ):
        self.key: Optional[Key] = key
        self.shell: Optional[ShellQuery] = shell
        self.counter = counter
        self.mode = mode or 'readable'
        self.block_id = block_id or 'head'
        self.address = address
        self.balance = balance
        self.amount = amount
        self.now = now
        self.level = level
        self.sender = sender
        self.source = source
        self.chain_id = chain_id
        self.protocol = protocol
        self.voting_power = voting_power
        self.total_voting_power = total_voting_power
        self.min_block_time = min_block_time
        self.tzt = tzt
        self.parameter_expr = get_script_section(script, name='parameter') if script and not tzt else None
        self.storage_expr = get_script_section(script, name='storage') if script and not tzt else None
        self.code_expr = get_script_section(script, name='code') if script else None
        self.views_expr = get_script_sections(script, name='view') if script else []
        self.input_expr = get_script_section(script, name='input') if script and tzt else None
        self.output_expr = get_script_section(script, name='output') if script and tzt else None
        self.sender_expr = get_script_section(script, name='sender') if script and tzt else None
        self.balance_expr = get_script_section(script, name='balance') if script and tzt else None
        self.amount_expr = get_script_section(script, name='amount') if script and tzt else None
        self.self_expr = get_script_section(script, name='self') if script and tzt else None
        self.now_expr = get_script_section(script, name='now') if script and tzt else None
        self.source_expr = get_script_section(script, name='source') if script and tzt else None
        self.chain_id_expr = get_script_section(script, name='chain_id') if script and tzt else None
        self.big_maps_expr = get_script_section(script, name='big_maps') if script and tzt else None
        self.origination_index = 1
        self.tmp_big_map_index = 0
        self.tmp_sapling_index = 0
        self.alloc_big_map_index = 0
        self.alloc_sapling_index = 0
        self.balance_update = 0
        self.big_maps = {}
        self.tzt_big_maps = {}
        self.view_results = view_results or {}
        self.global_constants = global_constants or {}
        self.debug = False
        self._sandboxed: Optional[bool] = None
        self.ipfs_gateway = (ipfs_gateway or DEFAULT_IPFS_GATEWAY).rstrip('/')
        self.storage_value = script.get('storage') if script else None

    def __copy__(self):
        raise ValueError("It's not allowed to copy context")

    @property
    def script(self) -> Optional[dict]:
        if self.parameter_expr and self.storage_expr and self.code_expr:
            return {
                'code': [
                    self.parameter_expr,
                    self.storage_expr,
                    self.code_expr,
                    *self.views_expr,
                ],
                'storage': self.storage_value,
            }
        else:
            return None

    @property
    def sandboxed(self) -> bool:
        if self.shell is None:
            raise Exception('`shell` is not set')
        if self._sandboxed is None:
            version = self.shell.version()
            self._sandboxed = version['network_version']['chain_name'] == 'SANDBOXED_TEZOS'
        return self._sandboxed

    def reset(self):
        self.counter = None
        self.origination_index = 1
        self.tmp_big_map_index = 0
        self.tmp_sapling_index = 0
        self.alloc_big_map_index = 0
        self.alloc_sapling_index = 0
        self.balance_update = 0
        self.big_maps.clear()
        self.tzt_big_maps.clear()
        self.global_constants.clear()

    def set_counter(self, counter: int):
        self.counter = counter

    def get_counter(self) -> int:
        if self.counter is None:
            if not self.key:
                raise Exception('key is undefined')
            if not self.shell:
                raise Exception('shell is undefined')

            key_hash = self.key.public_key_hash()
            self.counter = int(self.shell.contracts[key_hash]()['counter'])

        self.counter += 1
        return self.counter

    def get_counter_offset(self) -> int:
        """Return current count of pending transactions in mempool."""
        if self.key is None:
            raise Exception('`key` is not set')
        if self.shell is None:
            raise Exception('`shell` is not set')

        counter_offset = 0
        key_hash = self.key.public_key_hash()
        mempool = self.shell.mempool.pending_operations()

        for operation in chain(mempool.get('applied', []), mempool.get('unprocessed', [])):
            if isinstance(operation, list):
                operation = operation[1]
            for content in operation.get('contents', []):
                if content.get('source') == key_hash:
                    logger.debug("pending transaction in mempool: %s", content)
                    counter_offset += 1

        logger.debug("counter offset: %s", counter_offset)
        return counter_offset

    def register_big_map(self, ptr: int, copy=False) -> int:
        if copy:
            tmp_ptr = self.get_tmp_big_map_id()
            self.big_maps[tmp_ptr] = (ptr, True)
            return tmp_ptr
        else:
            self.big_maps[ptr] = (ptr, False)
            return ptr

    def get_tmp_big_map_id(self) -> int:
        self.tmp_big_map_index += 1
        return -self.tmp_big_map_index

    def get_big_map_diff(self, ptr: int) -> Tuple[Optional[int], int, str]:
        if ptr in self.big_maps:
            src_big_map, copy = self.big_maps[ptr]
            if copy:
                dst_big_map = self.alloc_big_map_index
                self.alloc_big_map_index += 1
                return src_big_map, dst_big_map, 'copy'
            else:
                return src_big_map, src_big_map, 'update'
        else:
            big_map = self.alloc_big_map_index
            self.alloc_big_map_index += 1
            return None, big_map, 'alloc'

    def get_originated_address(self) -> str:
        res = get_originated_address(self.origination_index)
        self.origination_index += 1
        return res

    def spend_balance(self, amount: int):
        balance = self.get_balance()
        assert amount <= balance, f'cannot spend {amount} tez, {balance} tez left'
        self.balance_update -= amount

    def get_parameter_expr(self, address=None) -> Optional[dict]:
        if self.shell and address:
            if address == get_originated_address(0):
                return None  # dummy callback
            else:
                script = self.shell.contracts[address].script()
                expr = get_script_section(script, name='parameter', cls=None, required=True)  # type: ignore
        elif address:
            return None
        else:
            expr = self.parameter_expr
        return self.resolve_global_constants(expr)

    def get_storage_expr(self, address=None) -> Optional[dict]:
        if self.shell and address:
            script = self.shell.contracts[address].script()
            expr = get_script_section(script, name='storage', cls=None, required=True)  # type: ignore
        elif address:
            return None
        else:
            expr = self.storage_expr
        return self.resolve_global_constants(expr)

    def get_storage_value(self, address=None) -> Optional[dict]:
        if self.shell:
            return self.shell.head.context.contracts[address].storage()
        return None if address else self.resolve_global_constants(self.storage_value)

    def get_code_expr(self):
        return self.resolve_global_constants(self.code_expr)

    def get_view_result(self, name, address=None) -> Optional[Any]:
        key = name if address is None else f'{address}%{name}'
        return self.view_results.get(key)

    def get_views_expr(self) -> List[dict]:
        return self.resolve_global_constants(self.views_expr)

    def get_view_expr(self, name, address=None) -> Optional[dict]:
        if address:
            if self.shell:
                script = self.shell.contracts[address].script()
                views = get_script_sections(script, name='view', cls=None)
            else:
                return None
        else:
            views = self.views_expr

        try:
            expr = next(view for view in views if view['args'][0]['string'] == name)
            return self.resolve_global_constants(expr)
        except (StopIteration, KeyError, IndexError):
            return None

    def get_input_expr(self):
        return self.input_expr

    def get_output_expr(self):
        return self.output_expr

    def get_sender_expr(self):
        return self.sender_expr

    def get_balance_expr(self):
        return self.balance_expr

    def get_amount_expr(self):
        return self.amount_expr

    def get_self_expr(self):
        return self.self_expr

    def get_now_expr(self):
        return self.now_expr

    def get_source_expr(self):
        return self.source_expr

    def get_chain_id_expr(self):
        return self.chain_id_expr

    def get_big_maps_expr(self):
        return self.big_maps_expr

    def set_storage_expr(self, expr):
        self.storage_expr = expr

    def set_parameter_expr(self, expr):
        self.parameter_expr = expr

    def set_code_expr(self, expr):
        self.code_expr = expr

    def set_input_expr(self, expr):
        self.input_expr = expr

    def set_output_expr(self, expr):
        self.output_expr = expr

    def set_source_expr(self, expr):
        self.source_expr = expr

    def set_chain_id_expr(self, expr):
        self.chain_id_expr = expr

    def set_big_maps_expr(self, expr):
        self.big_maps_expr = expr

    def get_big_map_value(self, ptr: int, key_hash: str):
        if self.tzt or (ptr not in self.big_maps):
            return None
        ptr, _ = self.big_maps[ptr]
        if ptr < 0:
            return None
        if self.shell is None:
            raise ValueError(f'Shell is undefined, cannot connect to network')
        try:
            return self.shell.blocks[self.block_id].context.big_maps[ptr][key_hash]()
        except RpcError:
            return None  # TODO: special exception/value | Key does not exist

    def register_sapling_state(self, ptr: int):
        raise NotImplementedError

    def get_tmp_sapling_state_id(self) -> int:
        self.tmp_sapling_index += 1
        return -self.tmp_sapling_index

    def get_sapling_state_diff(self, offset_commitment=0, offset_nullifier=0) -> Tuple[int, list]:
        ptr = self.alloc_sapling_index
        self.alloc_sapling_index += 1
        return ptr, []

    def get_self_address(self) -> str:
        return self.address or get_originated_address(0)

    def get_amount(self) -> int:
        return self.amount or 0

    def get_sender(self) -> str:
        return self.sender or self.get_dummy_key_hash()

    def get_source(self) -> str:
        return self.source or self.get_dummy_key_hash()

    def get_now(self) -> int:
        if self.now is not None:
            return self.now
        elif self.shell:
            ts = self.shell.head.header()['timestamp']
            dt = datetime.strptime(ts, '%Y-%m-%dT%H:%M:%SZ')
            first_delay = self.shell.head.context.constants().get('minimal_block_delay', 0)
            return int((dt - datetime(1970, 1, 1)).total_seconds()) + int(first_delay)
        else:
            return 0

    def get_level(self) -> int:
        if self.level is not None:
            return self.level
        elif self.shell:
            header = self.shell.blocks[self.block_id].header()
            return int(header['level'])
        else:
            return 1

    def get_balance(self) -> int:
        if self.balance is not None:
            balance = self.balance
        elif self.shell:
            contract = self.shell.contracts[self.get_self_address()]()
            balance = int(contract['balance'])
        else:
            balance = 0
        return balance + self.balance_update

    def get_voting_power(self, address: str) -> int:
        if self.voting_power is not None:
            return self.voting_power.get(address, 0)
        elif self.shell:
            raise NotImplementedError
        else:
            return 0

    def get_total_voting_power(self) -> int:
        if self.total_voting_power is not None:
            return self.total_voting_power
        elif self.shell:
            raise NotImplementedError
        else:
            return 0

    def get_min_block_time(self) -> int:
        if self.min_block_time:
            return self.min_block_time
        elif self.shell:
            constants = self.shell.head.context.constants()
            return int(constants['minimal_block_delay'])
        else:
            return 1

    def get_chain_id(self) -> str:
        if self.chain_id:
            return self.chain_id
        elif self.shell:
            return self.shell.chains.main.chain_id()
        else:
            return self.get_dummy_chain_id()

    def get_protocol(self) -> str:
        if self.protocol:
            return self.protocol
        elif self.shell:
            return self.shell.head.header()['protocol']
        else:
            raise NotImplementedError

    def get_dummy_address(self) -> str:
        if self.key:
            return self.key.public_key_hash()
        else:
            return base58_encode(b'\x00' * 20, b'KT1').decode()

    def get_dummy_txr_address(self) -> str:
        if self.key:
            return self.key.public_key_hash()
        else:
            return base58_encode(b'\x00' * 20, b'txr1').decode()

    def get_dummy_public_key(self) -> str:
        if self.key:
            return self.key.public_key()
        else:
            return base58_encode(b'\x00' * 32, b'edpk').decode()

    def get_dummy_key_hash(self) -> str:
        if self.key:
            return self.key.public_key_hash()
        else:
            return base58_encode(b'\x00' * 20, b'tz1').decode()

    def get_dummy_signature(self) -> str:
        return base58_encode(b'\x00' * 64, b'sig').decode()

    def get_dummy_chain_id(self) -> str:
        return base58_encode(b'\x00' * 4, b'Net').decode()

    def get_dummy_lambda(self):
        return {'prim': 'FAILWITH'}

    def set_total_voting_power(self, total_voting_power: int):
        self.total_voting_power = total_voting_power

    def set_voting_power(self, address: str, voting_power: int):
        self.voting_power[address] = voting_power

    def get_operations_ttl(self) -> int:
        if self.sandboxed:
            return MAX_OPERATIONS_TTL
        return DEFAULT_OPERATIONS_TTL

    def register_global_constant(self, expression):
        """Register global constant
        :param expression: Micheline expression
        """
        constant_hash = forge_script_expr(forge_micheline(expression))
        self.global_constants[constant_hash] = expression

    def resolve_global_constants(self, expression):
        """Replace global constants with their respectful values or throw an error if the constant is not defined
        :param expression: Micheline expression
        """

        def _resolve_constant(node):
            try:
                constant_hash = node['args'][0]['string']
            except (KeyError, IndexError) as e:
                raise ValueError('Unexpected constant expression') from e
            if constant_hash not in self.global_constants:
                raise KeyError(f'Constant {constant_hash} is not defined')
            return _resolve(self.global_constants[constant_hash])
            # TODO: check if global constants are really recursive

        def _resolve(node):
            if isinstance(node, dict):
                if node.get('prim') == 'constant':
                    return _resolve_constant(node)
                elif node.get('args'):
                    args = list(map(_resolve, node['args']))
                    return {k: v if k != 'args' else args for k, v in node.items()}
                else:
                    return node
            elif isinstance(node, list):
                return list(map(_resolve, node))
            else:
                return node

        return _resolve(expression)



import difflib
import re
from os.path import dirname
from os.path import join

import simplejson as json

# Based on https://gist.github.com/noporpoise/16e731849eb1231e86d78f9dfeca3abc
_no_eol = r'\ No newline at end of file'
_hdr_pat = re.compile(r'^@@ -(\d+),?(\d+)? \+(\d+),?(\d+)? @@$')
proj_dir = dirname(dirname(dirname(__file__)))


def make_patch(a, b, filename, context_size=0):
    """
    Get unified string diff between two strings. Trims top two lines.
    Returns empty string if strings are identical.
    """
    diffs = difflib.unified_diff(
        a=a.splitlines(True),
        b=b.splitlines(True),
        fromfile=filename,
        tofile=filename,
        n=context_size,
    )
    diffs = map(lambda x: x if x[-1] == '\n' else x + '\n' + _no_eol + '\n', diffs)
    return ''.join(diffs)


def apply_patch(source, patch, revert=False):
    """
    Apply patch to string s to recover newer string.
    If revert is True, treat s as the newer string, recover older string.
    """
    source = source.splitlines(True)
    patch = patch.splitlines(True)
    target = ''

    i = sl = 0
    (midx, sign) = (1, '+') if not revert else (3, '-')

    while i < len(patch) and patch[i].startswith(("---", "+++")):
        i += 1  # skip filename header

    while i < len(patch):
        match = _hdr_pat.match(patch[i])
        if not match:
            raise ValueError(f'Regex mismatch on line {i}, `{patch[i]}`')

        l = int(match.group(midx)) - 1 + (match.group(midx + 1) == '0')
        if sl > l or l > len(source):
            raise ValueError(f'Bad line num {i}: `{patch[i]}`')

        target += ''.join(source[sl:l])
        sl = l
        i += 1

        while i < len(patch) and patch[i][0] != '@':
            if i + 1 < len(patch) and patch[i + 1][0] == '\\':
                line = patch[i][:-1]
                i += 2
            else:
                line = patch[i]
                i += 1

            if len(line) > 0:
                if line[0] == sign or line[0] == ' ':
                    target += line[1:]
                sl += line[0] != sign

    target += ''.join(source[sl:])
    return target


def read_template(filename):
    with open(join(proj_dir, 'assets', filename)) as f:
        return f.read()


def generate_unidiff_html(diffs: list, output_path=None):
    html = read_template('unidiff.html')
    html = html.replace('{{text}}', json.dumps('\n'.join(diffs)))
    if output_path:
        with open(output_path, 'w') as f:
            f.write(html)
        return None
    return html


def generate_jsondiff_html(left: dict, right: dict, output_path=None):
    html = read_template('jsondiff.html')
    html = html.replace('{{left}}', json.dumps(left))
    html = html.replace('{{right}}', json.dumps(right))
    if output_path:
        with open(output_path, 'w') as f:
            f.write(html)
        return None
    return html



from copy import copy
from typing import List
from typing import Optional
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import parse_micheline_literal
from pytezos.michelson.types.base import MichelsonType


class SaplingTransactionType(MichelsonType, prim='sapling_transaction', args_len=1):
    pass


class SaplingTransactionDeprecatedType(MichelsonType, prim='sapling_transaction_deprecated', args_len=1):
    pass


class SaplingStateType(MichelsonType, prim='sapling_state', args_len=1):
    def __init__(self, ptr: Optional[int] = None):
        super(SaplingStateType, self).__init__()
        self.ptr = ptr
        self.context: Optional[AbstractContext] = None

    def __repr__(self):
        if self.ptr:
            return f'<{self.ptr}>'
        else:
            return '{}'

    @staticmethod
    def empty(memo_size: int):
        cls = SaplingStateType.create_type(args=[MichelineLiteral.create(memo_size)])
        return cls()

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'SaplingStateType':
        if isinstance(val_expr, dict):
            ptr = parse_micheline_literal(val_expr, {'int': int})
            return cls(ptr=ptr)
        elif isinstance(val_expr, list):
            return cls()
        else:
            raise AssertionError(f'unexpected value {val_expr}')

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if lazy_diff:
            return []
        else:
            assert self.ptr is not None, f'Sapling_state id is not defined'
            return {'int': str(self.ptr)}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        if self.ptr is not None:
            return self.ptr
        else:
            return []

    def to_literal(self) -> Type[Micheline]:
        if self.ptr is not None:
            return MichelineLiteral.create(self.ptr)
        else:
            return MichelineSequence.create_type(args=[])

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        self.context = context
        self.ptr = context.get_tmp_sapling_state_id()

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'MichelsonType':
        return copy(self)

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable') -> 'MichelsonType':
        assert self.ptr is not None, f'Sapling_state ID is not defined'
        if self.context:
            dst_ptr, _ = self.context.get_sapling_state_diff()
        else:
            dst_ptr = self.ptr
        return type(self)(dst_ptr)



from typing import List
from typing import Tuple
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import MichelsonType
from pytezos.michelson.types import OrType
from pytezos.michelson.types import PairType


def execute_cxr(prim: str, stack: MichelsonStack, stdout: List[str], idx: int):
    pair = cast(PairType, stack.pop1())
    pair.assert_type_in(PairType)
    res = pair.items[idx]
    stack.push(res)
    stdout.append(format_stdout(prim, [pair], [res]))


class CarInstruction(MichelsonInstruction, prim='CAR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_cxr(cls.prim, stack, stdout, 0)  # type: ignore
        return cls(stack_items_added=1)


class CdrInstruction(MichelsonInstruction, prim='CDR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_cxr(cls.prim, stack, stdout, 1)  # type: ignore
        return cls(stack_items_added=1)


class GetnInstruction(MichelsonInstruction, prim='GET', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        pair = cast(PairType, stack.pop1())
        pair.assert_type_in(PairType)
        index = cls.args[0].get_int()  # type: ignore
        res = pair.access_comb(index)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [pair], [res], index))  # type: ignore
        return cls(stack_items_added=1)


class UpdatenInstruction(MichelsonInstruction, prim='UPDATE', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        element, pair = cast(Tuple[MichelsonType, PairType], stack.pop2())
        pair.assert_type_in(PairType)
        index = cls.args[0].get_int()  # type: ignore
        res = pair.update_comb(index, element)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [element, pair], [res], index))  # type: ignore
        return cls(stack_items_added=1)


class LeftInstruction(MichelsonInstruction, prim='LEFT', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        left = stack.pop1()
        res = OrType.from_left(left, cls.args[0])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [left], [res]))  # type: ignore
        return cls()


class RightInstruction(MichelsonInstruction, prim='RIGHT', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        right = stack.pop1()
        res = OrType.from_right(right, cls.args[0])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [right], [res]))  # type: ignore
        return cls(stack_items_added=1)


class PairInstruction(MichelsonInstruction, prim='PAIR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        left, right = stack.pop2()
        res = PairType.from_comb([left, right])
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [left, right], [res]))  # type: ignore
        return cls(stack_items_added=1)


class UnpairInstruction(MichelsonInstruction, prim='UNPAIR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        pair = cast(PairType, stack.pop1())
        pair.assert_type_in(PairType)
        left, right = tuple(iter(pair))
        stack.push(right)
        stack.push(left)
        stdout.append(format_stdout(cls.prim, [pair], [left, right]))  # type: ignore
        return cls(stack_items_added=2)


class PairnInstruction(MichelsonInstruction, prim='PAIR', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        count = cls.args[0].get_int()  # type: ignore
        assert count >= 2, f'invalid argument, must be >= 2'
        leaves = stack.pop(count=count)
        res = PairType.from_comb(leaves)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, leaves, [res], count))  # type: ignore
        return cls(stack_items_added=1)


class UnpairnInstruction(MichelsonInstruction, prim='UNPAIR', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        count = cls.args[0].get_int()  # type: ignore
        assert count >= 2, f'invalid argument, must be >= 2'
        pair = cast(PairType, stack.pop1())
        pair.assert_type_in(PairType)
        leaves = list(pair.unpairn_comb(count - 2))
        for leaf in reversed(leaves):
            stack.push(leaf)
        stdout.append(format_stdout(cls.prim, [pair], leaves, count))  # type: ignore
        return cls(stack_items_added=len(leaves))



import inspect
import re
import sys
from functools import lru_cache
from functools import update_wrapper
from typing import Optional


@lru_cache(maxsize=1)
def is_interactive() -> bool:
    try:
        _ = get_ipython().__class__.__name__  # type: ignore
        return True
    except NameError:
        return any(lib in sys.modules for lib in ['pytest', 'unittest', 'nose'])


def get_attr_docstring(class_type, attr_name) -> Optional[str]:
    if attr_name == 'get':
        attr_name = '__call__'

    attr = getattr(class_type, attr_name, None)
    if attr and attr.__doc__:
        return re.sub(r' {3,}', '', attr.__doc__)
    return None


def default_attr_filter(x) -> bool:  # pylint: disable=unused-argument
    return True


def get_class_docstring(class_type, attr_filter=default_attr_filter, extended=False):
    def format_attribute(x):
        attr = getattr(class_type, x)
        if isinstance(attr, property):
            name = f'.{x}'
        else:
            if extended:
                sig = str(inspect.signature(attr)).replace('self, ', '')
            else:
                sig = '()'
            name = f'.{x}{sig}'

        if extended:
            doc = get_attr_docstring(class_type, x)
        else:
            doc = ''
        return f'{name}{doc}'

    def filter_attribute(x):
        return all(
            [
                not x.startswith('_'),
                attr_filter(x),
                not isinstance(getattr(class_type, x), property),
            ]
        )

    return '\n'.join(
        map(
            format_attribute,
            filter(
                filter_attribute,
                dir(class_type),
            ),
        ),
    )


def inline_doc(method):
    if not is_interactive():
        return method

    doc = [repr(method)]
    if method.__doc__:
        doc.append(re.sub(r' {3,}', '', method.__doc__))

    class CustomReprDescriptor:
        def __get__(self, obj, objtype=None):
            class MethodWrapper:
                def __init__(self):
                    self.obj = obj
                    self.doc = '\n'.join(doc)

                def __call__(self, *args, **kwargs):
                    if self.obj is None:
                        # it wasn't the case prior to 3.10: https://bugs.python.org/issue43682
                        return method(*args, **kwargs)
                    else:
                        return method(self.obj, *args, **kwargs)

                def __repr__(self):
                    return self.doc

            return update_wrapper(MethodWrapper(), method)

    return CustomReprDescriptor()


class InlineDocstring(type):
    def __new__(mcs, name, bases, attrs, **kwargs):
        if is_interactive():
            new_attrs = {}
            for attr_name, attr in attrs.items():
                if callable(attr) and attr.__doc__ and not attr_name.startswith('_'):
                    attr = inline_doc(attr)
                new_attrs[attr_name] = attr
        else:
            new_attrs = attrs
        return type.__new__(mcs, name, bases, new_attrs, **kwargs)



from typing import Any
from typing import Dict

from pytezos.michelson.forge import forge_address
from pytezos.michelson.forge import forge_array
from pytezos.michelson.forge import forge_base58
from pytezos.michelson.forge import forge_bool
from pytezos.michelson.forge import forge_int
from pytezos.michelson.forge import forge_int16
from pytezos.michelson.forge import forge_int32
from pytezos.michelson.forge import forge_micheline
from pytezos.michelson.forge import forge_nat
from pytezos.michelson.forge import forge_public_key
from pytezos.michelson.forge import forge_script
from pytezos.rpc.kind import operation_tags

reserved_entrypoints = {
    'default': b'\x00',
    'root': b'\x01',
    'do': b'\x02',
    'set_delegate': b'\x03',
    'remove_delegate': b'\x04',
}


def has_parameters(content: Dict[str, Any]) -> bool:
    if content.get('parameters'):
        if content['parameters']['entrypoint'] == 'default' and content['parameters']['value'] == {'prim': 'Unit'}:
            return False
        return True
    return False


def forge_entrypoint(entrypoint) -> bytes:
    """Encode Michelson contract entrypoint into the byte form.

    :param entrypoint: string
    """
    if entrypoint in reserved_entrypoints:
        return reserved_entrypoints[entrypoint]
    else:
        return b'\xff' + forge_array(entrypoint.encode(), len_bytes=1)


def forge_operation(content: Dict[str, Any]) -> bytes:
    """Forge operation content (locally).

    :param content: {.., "kind": "transaction", ...}
    """
    encode_content = {
        'failing_noop': forge_failing_noop,
        'activate_account': forge_activate_account,
        'reveal': forge_reveal,
        'transaction': forge_transaction,
        'origination': forge_origination,
        'delegation': forge_delegation,
        'endorsement': forge_endorsement,
        'endorsement_with_slot': forge_endorsement_with_slot,
        'register_global_constant': forge_register_global_constant,
    }
    encode_proc = encode_content.get(content['kind'])
    if not encode_proc:
        raise NotImplementedError(content['kind'])

    return encode_proc(content)


def forge_operation_group(operation_group: Dict[str, Any]) -> bytes:
    """Forge operation group (locally).

    :param operation_group: {"branch": "B...", "contents": [], ...}
    """
    res = forge_base58(operation_group['branch'])
    res += b''.join(map(forge_operation, operation_group['contents']))
    return res


def forge_activate_account(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_base58(content['pkh'])
    res += bytes.fromhex(content['secret'])
    return res


def forge_reveal(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_public_key(content['public_key'])
    return res


def forge_transaction(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_nat(int(content['amount']))
    res += forge_address(content['destination'])

    if has_parameters(content):
        res += forge_bool(True)
        res += forge_entrypoint(content['parameters']['entrypoint'])
        res += forge_array(forge_micheline(content['parameters']['value']))
    else:
        res += forge_bool(False)

    return res


def forge_origination(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_nat(int(content['balance']))

    if content.get('delegate'):
        res += forge_bool(True)
        res += forge_address(content['delegate'], tz_only=True)
    else:
        res += forge_bool(False)

    res += forge_script(content['script'])

    return res


def forge_delegation(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))

    if content.get('delegate'):
        res += forge_bool(True)
        res += forge_address(content['delegate'], tz_only=True)
    else:
        res += forge_bool(False)

    return res


def forge_endorsement(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_int32(int(content['level']))
    return res


def forge_inline_endorsement(content: Dict[str, Any]) -> bytes:
    res = forge_base58(content['branch'])
    res += forge_nat(operation_tags[content['operations']['kind']])
    res += forge_int32(int(content['operations']['level']))
    res += forge_base58(content['signature'])
    return res


def forge_endorsement_with_slot(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_array(forge_inline_endorsement(content['endorsement']))
    res += forge_int16(content['slot'])
    return res


def forge_failing_noop(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_array(content['arbitrary'].encode())
    return res


def forge_register_global_constant(content: Dict[str, Any]) -> bytes:
    res = forge_nat(operation_tags[content['kind']])
    res += forge_address(content['source'], tz_only=True)
    res += forge_nat(int(content['fee']))
    res += forge_nat(int(content['counter']))
    res += forge_nat(int(content['gas_limit']))
    res += forge_nat(int(content['storage_limit']))
    res += forge_array(forge_micheline(content['value']))
    return res



from typing import Callable
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import parse_micheline_value
from pytezos.michelson.types.base import MichelsonType


class EltLiteral(Micheline, prim='Elt', args_len=2):
    pass


class MapType(MichelsonType, prim='map', args_len=2):
    def __init__(self, items: List[Tuple[MichelsonType, MichelsonType]]):
        super(MapType, self).__init__()
        self.items = items

    def __repr__(self):
        elements = [f'{repr(k)}: {repr(v)}' for k, v in self.items]
        return f'{{{", ".join(elements)}}}'

    def __len__(self):
        return len(self.items)

    def __iter__(self) -> Generator[Tuple[MichelsonType, MichelsonType], None, None]:
        yield from iter(self.items)

    def __eq__(self, other) -> bool:
        if not isinstance(other, MapType):
            return False
        return self.items == other.items

    @staticmethod
    def empty(key_type: Type[MichelsonType], val_type: Type[MichelsonType]) -> 'MapType':
        cls = MapType.create_type(args=[key_type, val_type])
        return cls(items=[])  # type: ignore

    @staticmethod
    def from_items(items: List[Tuple[MichelsonType, MichelsonType]]) -> 'MapType':
        assert len(items) > 0, 'cannot instantiate from empty list'
        key_type, val_type = items[0][0].get_anon_type(), items[0][1].get_anon_type()
        for key, val in items[1:]:
            key_type.assert_type_equal(type(key))
            val_type.assert_type_equal(type(val))
        cls = MapType.create_type(args=[key_type, val_type])
        cls.check_constraints(items)  # type: ignore
        return cls(items=items)  # type: ignore

    @classmethod
    def check_constraints(cls, items: List[Tuple[MichelsonType, MichelsonType]]):
        keys = list(map(lambda x: x[0], items))
        assert len(set(keys)) == len(keys), f'duplicate keys found'
        assert keys == sorted(keys), f'keys are unsorted'

    @classmethod
    def generate_pydoc(cls, definitions: List[Tuple[str, str]], inferred_name=None, comparable=False):
        name = cls.field_name or cls.type_name or inferred_name
        key = cls.args[0].generate_pydoc(
            definitions,
            inferred_name=f'{name}_key' if name else None,
            comparable=True,
        )
        val = cls.args[1].generate_pydoc(
            definitions,
            inferred_name=f'{name}_value' if name else None,
        )
        return f'{{ {key}: {val}, … }}'

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'MapType':
        return cls([])

    @classmethod
    def parse_micheline_value(cls, val_expr) -> List[Tuple[MichelsonType, MichelsonType]]:
        assert isinstance(val_expr, list), f'expected list, got {type(val_expr).__name__}'

        def parse_elt(elt_expr) -> Tuple[MichelsonType, MichelsonType]:
            return parse_micheline_value(
                elt_expr,
                {('Elt', 2): lambda x: tuple(cls.args[i].from_micheline_value(arg) for i, arg in enumerate(x))},
            )

        items = list(map(parse_elt, val_expr))
        cls.check_constraints(items)
        return items

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'MapType':
        return cls(cls.parse_micheline_value(val_expr))

    @classmethod
    def parse_python_object(cls, py_obj) -> List[Tuple[MichelsonType, MichelsonType]]:
        assert isinstance(py_obj, dict), f'expected dict, got {type(py_obj).__name__}'
        items = [(cls.args[0].from_python_object(k), cls.args[1].from_python_object(v)) for k, v in py_obj.items()]
        return sorted(items, key=lambda x: x[0])

    @classmethod
    def from_python_object(cls, py_obj) -> 'MapType':
        return cls(cls.parse_python_object(py_obj))

    def to_literal(self) -> Type[Micheline]:
        return MichelineSequence.create_type(
            args=[
                EltLiteral.create_type(
                    args=[
                        k.to_literal(),
                        v.to_literal(),
                    ]
                )
                for k, v in self.items
            ]
        )

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return [
            {
                'prim': 'Elt',
                'args': [x.to_micheline_value(mode=mode, lazy_diff=lazy_diff) for x in elt],
            }
            for elt in self
        ]

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False) -> dict:
        assert not comparable, f'{self.prim} is not comparable'
        return {
            k.to_python_object(try_unpack=try_unpack, comparable=True): v.to_python_object(
                try_unpack=try_unpack,
                lazy_diff=lazy_diff,
            )
            for k, v in self.items
        }

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'MapType':
        items = [(key, val.merge_lazy_diff(lazy_diff)) for key, val in self.items]
        return type(self)(items)

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable'):
        items = [(key, val.aggregate_lazy_diff(lazy_diff, mode=mode)) for key, val in self.items]
        return type(self)(items)

    def find(self, predicate: Callable[['MichelsonType'], bool]) -> Optional['MichelsonType']:
        if predicate(self):
            return self
        for _, item in self:
            res = item.find(predicate)
            if res:
                return res
        return None

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        for _, val in self.items:
            val.attach_context(context, big_map_copy=big_map_copy)

    def get(self, key: MichelsonType, dup=True) -> Optional[MichelsonType]:
        self.args[0].assert_type_equal(type(key))
        if dup:
            assert self.args[1].is_duplicable(), f'use GET_AND_UPDATE instead'
        return next((v for k, v in self.items if k == key), None)

    def contains(self, key: MichelsonType):
        return self.get(key, dup=False) is not None

    def update(self, key: MichelsonType, val: Optional[MichelsonType]) -> Tuple[Optional[MichelsonType], MichelsonType]:
        prev_val = self.get(key, dup=False)
        if prev_val is not None:
            if val is not None:
                items = [(k, v if k != key else val) for k, v in self.items]
            else:  # remove
                items = [(k, v) for k, v in self.items if k != key]
        else:
            if val is not None:
                items = sorted(self.items + [(key, val)], key=lambda x: x[0])
            else:  # do nothing
                items = self.items
        return prev_val, type(self)(items)

    def __contains__(self, key_obj):
        key = self.args[0].from_python_object(key_obj)
        return self.contains(key)

    def __getitem__(self, key_obj) -> MichelsonType:
        key = self.args[0].from_python_object(key_obj)
        val = self.get(key)
        assert val is not None, f'not found: {key_obj}'
        return val



from typing import Any
from typing import List
from typing import Optional
from typing import Type
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.types.base import MichelsonType


class ViewSection(Micheline, prim='view', args_len=4):
    """
    Syntax: view {name} {arg_type} {ret_type} {code}
    """

    args: List[Type[MichelsonType]]  # type: ignore
    name: str

    def __init__(self, item: MichelsonType):
        super().__init__()
        self.item = item

    def __repr__(self):
        return repr(self.item)

    @staticmethod
    def match(type_expr) -> Type['ViewSection']:
        try:
            cls = Micheline.match(type_expr)
            if not issubclass(cls, ViewSection):
                cls = ViewSection.create_type(args=[cls])
        except Exception as e:
            raise MichelsonRuntimeError('view', *e.args) from e
        return cls

    @staticmethod
    def check_code(code: Type['Micheline'], lambda_: bool) -> None:
        if code.prim == 'SELF':
            raise MichelsonRuntimeError('view', f'{code.prim} is not allowed in views')
        if code.prim in ('CREATE_CONTRACT', 'SET_DELEGATE', 'TRANSFER_TOKENS') and not lambda_:
            raise MichelsonRuntimeError('view', f'{code.prim} is not allowed in views')

        lambda_ |= code.prim in ('LAMBDA', 'lambda')
        for arg in getattr(code, 'args', ()):
            ViewSection.check_code(arg, lambda_)

    @classmethod
    def create_type(
        cls,
        args: List[Union[Type['Micheline'], Any]],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['ViewSection']:
        view_name = cast(Type[MichelineLiteral], args[0])
        if not issubclass(view_name, MichelineLiteral):
            raise MichelsonRuntimeError('view', 'Expected view name as first argument', view_name)
        name = view_name.get_string()
        if len(name) >= 32:
            # TODO: also check for denied symbols
            raise MichelsonRuntimeError('view', f'Too long view name {view_name}')

        # NOTE: Check for opcodes forbidden in views
        cls.check_code(args[3], lambda_=False)

        res = type(cls.__name__, (cls,), dict(args=args, name=name, **kwargs))
        return cast(Type['ViewSection'], res)

    @classmethod
    def generate_pydoc(cls) -> str:
        definitions = []  # type: ignore
        return cls.args[1].generate_pydoc(definitions, cls.prim)



from decimal import Decimal
from typing import Type
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.context.abstract import get_originated_address
from pytezos.crypto.encoding import base58_decode
from pytezos.crypto.encoding import is_address
from pytezos.crypto.encoding import is_chain_id
from pytezos.crypto.encoding import is_kt
from pytezos.crypto.encoding import is_pkh
from pytezos.crypto.encoding import is_public_key
from pytezos.crypto.encoding import is_sig
from pytezos.crypto.encoding import is_txr_address
from pytezos.michelson.forge import forge_address
from pytezos.michelson.forge import forge_base58
from pytezos.michelson.forge import forge_contract
from pytezos.michelson.forge import forge_public_key
from pytezos.michelson.forge import optimize_timestamp
from pytezos.michelson.forge import unforge_address
from pytezos.michelson.forge import unforge_chain_id
from pytezos.michelson.forge import unforge_contract
from pytezos.michelson.forge import unforge_public_key
from pytezos.michelson.forge import unforge_signature
from pytezos.michelson.format import format_timestamp
from pytezos.michelson.format import micheline_to_michelson
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import parse_micheline_literal
from pytezos.michelson.parse import michelson_to_micheline
from pytezos.michelson.types.base import Undefined
from pytezos.michelson.types.core import IntType
from pytezos.michelson.types.core import MichelsonType
from pytezos.michelson.types.core import NatType
from pytezos.michelson.types.core import StringType


class TimestampType(IntType, prim='timestamp'):  # type: ignore
    @classmethod
    def from_value(cls, value: int) -> 'TimestampType':
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'TimestampType':
        value = parse_micheline_literal(val_expr, {'int': int, 'string': optimize_timestamp})
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'TimestampType':
        if isinstance(py_obj, int):
            value = py_obj
        elif isinstance(py_obj, str):
            value = optimize_timestamp(py_obj)
        else:
            raise AssertionError(f'unexpected value type {py_obj}')
        return cls.from_value(value)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'int': str(self.value)}
        elif mode == 'readable':
            return {'string': format_timestamp(self.value)}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class MutezType(NatType, prim='mutez'):
    def __repr__(self):
        return str(Decimal(self.value) / 10**6)

    @classmethod
    def from_value(cls, value: int) -> 'MutezType':
        assert value >= 0, f'expected natural number, got {value}'
        if value.bit_length() > 63:
            raise OverflowError(f'mutez overflow, got {value.bit_length()} bits, should not exceed 63')
        return cls(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'MutezType':
        if isinstance(py_obj, int):
            value = py_obj
        elif isinstance(py_obj, Decimal):
            value = int(py_obj * (10**6))
        elif isinstance(py_obj, str):
            value = int(Decimal(py_obj) * (10**6))
        else:
            raise AssertionError(f'unexpected value type {py_obj}')
        return cls.from_value(value)


class AddressType(StringType, prim='address'):
    def __repr__(self):
        return f'{self.value[:6]}…{self.value[-3:]}'

    def __lt__(self, other: 'AddressType') -> bool:  # type: ignore
        if is_pkh(self.value) and is_kt(other.value):
            return True
        elif is_kt(self.value) and is_pkh(other.value):
            return False
        else:
            return self.value < other.value

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'AddressType':
        return cls.from_value(context.get_dummy_address())

    @classmethod
    def from_value(cls, value: str) -> 'AddressType':
        if value.endswith('%default'):
            value = value.split('%')[0]
        assert is_address(value), f'expected tz/KT address, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'AddressType':
        value = parse_micheline_literal(
            val_expr,
            {
                'bytes': lambda x: unforge_contract(bytes.fromhex(x)),
                'string': lambda x: x,
            },
        )
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'AddressType':
        return cls.from_value(py_obj)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'bytes': forge_contract(self.value).hex()}  # because address can also have an entrypoint
        elif mode == 'readable':
            return {'string': self.value}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class TXRAddress(StringType, prim='tx_rollup_l2_address'):
    def __repr__(self):
        return f'{self.value[:6]}…{self.value[-3:]}'

    def __lt__(self, other: 'TXRAddress') -> bool:  # type: ignore
        if is_kt(other.value):
            return True
        elif is_pkh(other.value):
            return False
        else:
            return self.value < other.value

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'TXRAddress':
        return cls.from_value(context.get_dummy_txr_address())

    @classmethod
    def from_value(cls, value: str) -> 'TXRAddress':
        if value.endswith('%default'):
            value = value.split('%')[0]
        assert is_txr_address(value), f'expected txr1 address, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'TXRAddress':
        value = parse_micheline_literal(
            val_expr,
            {
                'bytes': lambda x: unforge_contract(bytes.fromhex(x)),
                'string': lambda x: x,
            },
        )
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'TXRAddress':
        return cls.from_value(py_obj)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'bytes': forge_contract(self.value).hex()}  # because address can also have an entrypoint
        elif mode == 'readable':
            return {'string': self.value}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class KeyType(StringType, prim='key'):
    @property
    def raw(self) -> bytes:
        return base58_decode(self.value.encode())

    @property
    def prefix(self) -> str:
        return self.value[:4]

    def __lt__(self, other: 'KeyType') -> bool:  # type: ignore
        """
        Keys are ordered as follows: edpk < sppk < p2pk
        All keys are in compressed form in Tezos (flag | X) where flag specifies if Y is odd or even
        https://crypto.stackexchange.com/questions/70754/ec-key-compression
        For secp256r1 (aka p256) we need to cut the first byte (for unknown reason)
        """
        curves = {
            'edpk': (0, 0),
            'sppk': (1, 0),
            'p2pk': (2, 1),
        }
        res = curves[self.prefix][0] - curves[other.prefix][0]
        if res < 0:
            return True
        elif res > 0:
            return False
        else:
            offset = curves[self.prefix][1]
            return self.raw[offset:] < other.raw[offset:]

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'KeyType':
        return cls.from_value(context.get_dummy_public_key())

    @classmethod
    def from_value(cls, value: str) -> 'KeyType':
        assert is_public_key(value), f'expected ed/sp/p2 public key, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'KeyType':
        value = parse_micheline_literal(
            val_expr,
            {
                'bytes': lambda x: unforge_public_key(bytes.fromhex(x)),
                'string': lambda x: x,
            },
        )
        return cls.from_value(value)

    @classmethod
    def parse_python_object(cls, py_obj) -> 'KeyType':
        return cls.from_value(py_obj)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'bytes': forge_public_key(self.value).hex()}
        elif mode == 'readable':
            return {'string': self.value}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class KeyHashType(StringType, prim='key_hash'):
    @classmethod
    def dummy(cls, context: AbstractContext) -> 'KeyHashType':
        return cls.from_value(context.get_dummy_key_hash())

    @classmethod
    def from_value(cls, value: str) -> 'KeyHashType':
        assert is_pkh(value), f'expected tz1/tz2/tz3 key hash, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'KeyHashType':
        value = parse_micheline_literal(
            val_expr, {'bytes': lambda x: unforge_address(bytes.fromhex(x)), 'string': lambda x: x}
        )
        return cls.from_value(value)

    @classmethod
    def parse_python_object(cls, py_obj):
        return cls.from_value(py_obj)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'bytes': forge_address(self.value, tz_only=True).hex()}
        elif mode == 'readable':
            return {'string': self.value}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class SignatureType(StringType, prim='signature'):
    @classmethod
    def dummy(cls, context: AbstractContext) -> 'SignatureType':
        return cls.from_value(context.get_dummy_signature())

    @classmethod
    def from_value(cls, value: str) -> 'SignatureType':
        assert is_sig(value), f'expected signature, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'SignatureType':
        value = parse_micheline_literal(
            val_expr,
            {
                'bytes': lambda x: unforge_signature(bytes.fromhex(x)),
                'string': lambda x: x,
            },
        )
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'SignatureType':
        return cls.from_value(py_obj)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'bytes': forge_base58(self.value).hex()}
        elif mode == 'readable':
            return {'string': self.value}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class ChainIdType(StringType, prim='chain_id'):
    @classmethod
    def dummy(cls, context: AbstractContext) -> 'ChainIdType':
        return cls.from_value(context.get_dummy_chain_id())

    @classmethod
    def from_value(cls, value: str) -> 'ChainIdType':
        assert is_chain_id(value), f'expected chain id, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'ChainIdType':
        value = parse_micheline_literal(
            val_expr, {'bytes': lambda x: unforge_chain_id(bytes.fromhex(x)), 'string': lambda x: x}
        )
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'ChainIdType':
        return cls.from_value(py_obj)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if mode in ['optimized', 'legacy_optimized']:
            return {'bytes': forge_base58(self.value).hex()}
        elif mode == 'readable':
            return {'string': self.value}
        else:
            raise AssertionError(f'unsupported mode {mode}')

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class ContractType(AddressType, prim='contract', args_len=1):
    def __repr__(self):
        address, entrypoint = self.get_address(), self.get_entrypoint()
        return f'{address[:6]}…{address[-3:]}%{entrypoint}'

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        super(ContractType, cls).generate_pydoc(definitions)
        param_expr = micheline_to_michelson(cls.args[0].as_micheline_expr())
        if cls.args[0].args:
            name = cls.field_name or cls.type_name or inferred_name or f'{cls.prim}_{len(definitions)}'
            param_name = f'{name}_param'
            definitions.insert(0, (param_name, param_expr))
            return f'contract (${param_name})'
        else:
            return f'contract ({param_expr})'

    @classmethod
    def from_python_object(cls, py_obj) -> 'ContractType':
        if py_obj is None or py_obj is Undefined:
            py_obj = get_originated_address(0)
        res = super(ContractType, cls).from_python_object(py_obj)
        return cast(ContractType, res)

    def get_address(self) -> str:
        return self.value.split('%')[0]

    def get_entrypoint(self) -> str:
        res = self.value.split('%')
        return res[1] if len(res) == 2 else 'default'

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return super(ContractType, self).to_python_object()


class LambdaType(MichelsonType, prim='lambda', args_len=2):  # type: ignore
    def __init__(self, value: Type[Micheline]):
        super(LambdaType, self).__init__()
        self.value = value

    def __repr__(self):
        return 'Lambda'

    def __eq__(self, other) -> bool:
        if not isinstance(other, LambdaType):
            return False
        # FIXME: That seems ugly
        return self.value.as_micheline_expr() == other.value.as_micheline_expr()

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        super(LambdaType, cls).generate_pydoc(definitions)
        name = cls.field_name or cls.type_name or inferred_name or f'{cls.prim}_{len(definitions)}'
        expr = {}
        for i, suffix in enumerate(['return', 'param']):
            arg_expr = micheline_to_michelson(cls.args[i].as_micheline_expr())
            if cls.args[i].args:
                arg_name = f'{name}_{suffix}'
                definitions.insert(0, (arg_name, arg_expr))
                expr[suffix] = f'${arg_name}'
            else:
                expr[suffix] = arg_expr
        return f'lambda ({expr["param"]} -> {expr["return"]})'

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'LambdaType':
        return cls(Micheline.match(context.get_dummy_lambda()))

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'LambdaType':
        assert isinstance(val_expr, list), f'expected list, got {type(val_expr).__name__}'
        return cls(Micheline.match(val_expr))

    @classmethod
    def from_python_object(cls, py_obj) -> 'LambdaType':
        assert isinstance(py_obj, str), f'expected string, got {type(py_obj).__name__}'
        value = michelson_to_micheline(py_obj)
        return cls.from_micheline_value(value)

    def to_literal(self) -> Type[Micheline]:
        return self.value

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        # TODO: optimized mode -> harcoded values in the code
        return self.value.as_micheline_expr()

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return micheline_to_michelson(self.to_micheline_value())



from typing import List
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.types import *


class StorageSection(Micheline, prim='storage', args_len=1):
    args: List[Type[MichelsonType]]  # type: ignore

    def __init__(self, item: MichelsonType):
        super().__init__()
        self.item = item

    def __repr__(self):
        return repr(self.item)

    @staticmethod
    def match(type_expr) -> Type['StorageSection']:
        try:
            cls = Micheline.match(type_expr)
            if not issubclass(cls, StorageSection):
                cls = StorageSection.create_type(args=[cls])
            assert cls.args[0].field_name is None, f'argument type cannot be annotated: %{cls.args[0].field_name}'  # type: ignore
        except Exception as e:
            raise MichelsonRuntimeError('storage', *e.args) from e
        return cls  # type: ignore

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_storage_expr(cls.as_micheline_expr())
        stdout.append('storage: updated')

    @classmethod
    def generate_pydoc(cls) -> str:
        definitions = []  # type: ignore
        res = cls.args[0].generate_pydoc(definitions, cls.prim)
        if res != f'${cls.prim}':
            definitions.insert(0, (cls.prim, res))
        return '\n'.join(f'${var}:\n\t{doc}\n' for var, doc in definitions)

    @classmethod
    def dummy(cls, context: AbstractContext):
        return cls(cls.args[0].dummy(context))

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'StorageSection':
        item = cls.args[0].from_micheline_value(val_expr)
        return cls(item)

    @classmethod
    def from_python_object(cls, py_obj) -> 'StorageSection':
        item = cls.args[0].from_python_object(py_obj)
        return cls(item)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return self.item.to_micheline_value(mode=mode, lazy_diff=lazy_diff)

    def to_python_object(self, try_unpack=False, lazy_diff=False):
        return self.item.to_python_object(
            try_unpack=try_unpack,
            lazy_diff=lazy_diff,
        )

    def attach_context(self, context: AbstractContext):
        self.item.attach_context(context)

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'StorageSection':
        item = self.item.merge_lazy_diff(lazy_diff)
        return type(self)(item)

    def aggregate_lazy_diff(self, mode='readable') -> List[dict]:
        lazy_diff = []  # type: ignore
        self.item.aggregate_lazy_diff(lazy_diff, mode=mode)
        return lazy_diff



from typing import List
from typing import Tuple
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import MichelsonType
from pytezos.michelson.types import NatType
from pytezos.michelson.types import OptionType
from pytezos.michelson.types import PairType
from pytezos.michelson.types import TicketType


class JoinTicketsInstruction(MichelsonInstruction, prim='JOIN_TICKETS'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        pair = cast(PairType, stack.pop1())
        pair.assert_type_in(PairType)
        left, right = tuple(pair)
        assert isinstance(left, TicketType), f'expected ticket on the left, got {left.prim}'
        assert isinstance(right, TicketType), f'expected ticket on the right, got {right.prim}'
        res = TicketType.join(left, right)
        if res is None:
            res = OptionType.none(type(left))  # type: ignore
        else:
            res = OptionType.from_some(res)  # type: ignore
        stack.push(res)  # type: ignore
        stdout.append(format_stdout(cls.prim, [pair], [res]))  # type: ignore
        return cls(stack_items_added=1)


class ReadTicketInstruction(MichelsonInstruction, prim='READ_TICKET'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        ticket = cast(TicketType, stack.pop1())
        ticket.assert_type_in(TicketType)
        res = ticket.to_comb()
        stack.push(ticket)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [ticket], [res, ticket]))  # type: ignore
        return cls(stack_items_added=2)


class SplitTicketInstruction(MichelsonInstruction, prim='SPLIT_TICKET'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        ticket, amounts = cast(Tuple[TicketType, PairType], stack.pop2())
        ticket.assert_type_in(TicketType)
        amounts.assert_type_in(PairType)
        a, b = cast(Tuple[NatType, NatType], tuple(amounts))
        a.assert_type_equal(NatType)
        b.assert_type_equal(NatType)
        res = ticket.split(int(a), int(b))
        if res is None:
            res = OptionType.none(PairType.create_type(args=[type(ticket), type(ticket)]))  # type: ignore
        else:
            res = OptionType.from_some(PairType.from_comb(list(res)))  # type: ignore
        stack.push(res)  # type: ignore
        stdout.append(format_stdout(cls.prim, [ticket, amounts], [res]))  # type: ignore
        return cls(stack_items_added=1)


class TicketInstruction(MichelsonInstruction, prim='TICKET'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        item, amount = cast(Tuple[MichelsonType, NatType], stack.pop2())
        amount.assert_type_equal(NatType)
        address = context.get_self_address()
        res = TicketType.create(address, item, int(amount))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [item, amount], [res]))  # type: ignore
        return cls(stack_items_added=1)



from copy import deepcopy
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

from attr import dataclass

from pytezos.context.impl import ExecutionContext
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.parse import MichelsonParser
from pytezos.michelson.parse import MichelsonParserError
from pytezos.michelson.parse import michelson_to_micheline
from pytezos.michelson.program import MichelsonProgram
from pytezos.michelson.program import TztMichelsonProgram
from pytezos.michelson.sections import CodeSection
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import OperationType


@dataclass(kw_only=True)
class InterpreterResult:
    """Result of running contract in interpreter"""

    operations = None
    storage = None
    lazy_diff = None
    stdout: List[str]
    error: Optional[Exception] = None
    instructions: Optional[MichelineSequence] = None
    stack: Optional[MichelsonStack] = None


class Interpreter:
    """Michelson interpreter reimplemented in Python.
    Based on the following reference: https://tezos.gitlab.io/michelson-reference/
    """

    def __init__(
        self,
        extra_primitives: Optional[List[str]] = None,
        debug: bool = False,
    ) -> None:
        self.stack = MichelsonStack()
        self.context = ExecutionContext()
        self.context.debug = debug
        self.parser = MichelsonParser(debug=debug, extra_primitives=extra_primitives)

    def execute(self, code: str) -> InterpreterResult:
        """Execute some code preserving current context and stack

        :param code: Michelson code
        """
        result = InterpreterResult(stdout=[])
        stack_backup = deepcopy(self.stack)
        context_backup = deepcopy(self.context)

        try:
            code_section = CodeSection.match(michelson_to_micheline(code))
            instructions = code_section.args[0].execute(self.stack, result.stdout, self.context)
            result.instructions = MichelineSequence([instructions])
            result.stack = self.stack
        except (MichelsonParserError, MichelsonRuntimeError) as e:
            if self.context.debug:
                raise

            self.stack = stack_backup
            self.context = context_backup
            result.stdout.append(e.format_stdout())
            result.error = e

        return result

    def reset(self) -> None:
        """Reset interpreter's stack and context"""
        self.stack = MichelsonStack()
        self.context = ExecutionContext()

    @staticmethod
    def run_code(
        parameter,
        storage,
        script: str,
        entrypoint='default',
        output_mode='readable',
        amount=None,
        chain_id=None,
        source=None,
        sender=None,
        balance=None,
        block_id=None,
        **kwargs,
    ) -> Tuple[List[dict], Any, List[dict], List[str], Optional[Exception]]:
        """Execute contract in interpreter

        :param parameter: parameter expression
        :param storage: storage expression
        :param script: contract's Michelson code
        :param entrypoint: contract entrypoint
        :param output_mode: one of readable/optimized/legacy_optimized
        :param amount: patch AMOUNT
        :param chain_id: patch CHAIN_ID
        :param source: patch SOURCE
        :param sender: patch SENDER
        :param balance: patch BALANCE
        :param block_id: set block ID
        """
        context = ExecutionContext(
            amount=amount,
            chain_id=chain_id,
            source=source,
            sender=sender,
            balance=balance,
            block_id=block_id,
            script={'code': script, 'storage': storage},
            **kwargs,
        )
        stack = MichelsonStack()
        stdout = []  # type: ignore
        try:
            program = MichelsonProgram.load(context, with_code=True)
            res = program.instantiate(
                entrypoint=entrypoint,
                parameter=parameter,
                storage=storage,
            )
            res.begin(stack, stdout, context)
            res.execute(stack, stdout, context)
            operations, storage, lazy_diff, _ = res.end(stack, stdout, output_mode=output_mode)
            return operations, storage, lazy_diff, stdout, None
        except MichelsonRuntimeError as e:
            stdout.append(e.format_stdout())
            return [], None, [], stdout, e

    @staticmethod
    def run_callback(
        entrypoint: str,
        parameter,
        storage,
        context: ExecutionContext,
    ) -> Tuple[Any, Any, List[str], Optional[Exception]]:
        """Execute view entrypoint of the contract loaded into the context

        :param entrypoint: contract entrypoint
        :param parameter: parameter section
        :param storage: storage section
        :param context: execution context
        :returns: [operations, storage, stdout, error]
        """
        ctx = ExecutionContext(
            shell=context.shell,
            key=context.key,
            block_id=context.block_id,
            script=context.script,
            address=context.address,
        )
        stack = MichelsonStack()
        stdout = []  # type: ignore
        try:
            program = MichelsonProgram.load(ctx, with_code=True)
            res = program.instantiate(entrypoint=entrypoint, parameter=parameter, storage=storage)
            res.begin(stack, stdout, context)
            res.execute(stack, stdout, context)
            _, _, _, pair = res.end(stack, stdout)
            operations = cast(List[OperationType], list(pair.items[0]))
            storage = pair.items[1]
            # Note: the `storage` returned by the Michelson interpreter above is not
            # required to include the full annotations specified in the contract's storage.
            # The lack of annotations affects calls to `to_python_object()`, causing the storage
            # you get back from the view to not always be converted to the same object
            # as if you called ContractInterface.storage() directly.
            # Re-parsing using the contract's storage section here to recover the annotations.
            storage = program.storage.from_micheline_value(storage.to_micheline_value())
            return [op.to_python_object() for op in operations], storage.to_python_object(), stdout, None
        except MichelsonRuntimeError as e:
            stdout.append(e.format_stdout())
            return None, None, stdout, e

    @staticmethod
    def run_view(name: str, parameter, storage, context: ExecutionContext) -> Tuple[Any, Any, Optional[Exception]]:
        ctx = ExecutionContext(
            shell=context.shell,
            key=context.key,
            block_id=context.block_id,
            script=context.script,
            address=context.address,
            view_results=context.view_results,
        )
        stack = MichelsonStack()
        stdout = []  # type: ignore
        try:
            program = MichelsonProgram.load(ctx, with_code=True)
            res = program.instantiate_view(name=name, parameter=parameter, storage=storage)
            res.begin(stack, stdout, context)
            res.execute_view(stack, stdout, context)
            ret_value = res.ret(stack, stdout)
            return ret_value.to_python_object(), stdout, None
        except MichelsonRuntimeError as e:
            stdout.append(e.format_stdout())
            return None, stdout, e

    @staticmethod
    def run_tzt(
        script: str,
        amount=None,
        chain_id=None,
        source=None,
        sender=None,
        balance=None,
        block_id=None,
        **kwargs,
    ) -> None:
        """Execute TZT test suite code

        :param script: test contract's Michelson code
        :param amount: patch AMOUNT
        :param chain_id: patch CHAIN_ID
        :param source: patch SOURCE
        :param sender: patch SENDER
        :param balance: patch BALANCE
        :param block_id: set block ID
        """
        context = ExecutionContext(
            amount=amount,
            chain_id=chain_id,
            source=source,
            sender=sender,
            balance=balance,
            block_id=block_id,
            script={'code': script},
            tzt=True,
            **kwargs,
        )
        stack = MichelsonStack()
        stdout: List[str] = []

        program = TztMichelsonProgram.load(context, with_code=True)
        res = program.instantiate()
        res.fill_context(script, context)
        res.register_bigmaps(stack, stdout, context)
        res.begin(stack, stdout, context)
        res.execute(stack, stdout, context)
        res.end(stack, stdout, context)



from copy import copy
from pprint import pformat
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.format import micheline_to_michelson
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.domain import AddressType
from pytezos.michelson.types.domain import NatType
from pytezos.michelson.types.pair import PairType


class TicketType(MichelsonType, prim='ticket', args_len=1):
    def __init__(self, ticketer: str, item: MichelsonType, amount: int):
        super(TicketType, self).__init__()
        self.ticketer = ticketer
        self.item = item
        self.amount = amount

    def __copy__(self):
        raise AssertionError('forbidden')

    def __repr__(self):
        return pformat((self.ticketer, repr(self.item), self.amount))

    @staticmethod
    def create(ticketer: str, item: MichelsonType, amount: int) -> 'TicketType':
        cls = TicketType.create_type(args=[item.get_anon_type()])
        return cls(ticketer, item, amount)  # type: ignore

    @classmethod
    def from_comb(cls, comb: PairType) -> 'TicketType':
        ticketer, item, amount = cast(Tuple[AddressType, MichelsonType, NatType], tuple(comb.iter_comb()))
        return cls(
            item=item,
            ticketer=str(ticketer),
            amount=int(amount),
        )

    @staticmethod
    def join(left: 'TicketType', right: 'TicketType') -> Optional['TicketType']:
        left.assert_type_equal(type(right))
        if left.ticketer != right.ticketer or left.item != right.item:
            return None
        else:
            return TicketType(ticketer=left.ticketer, item=left.item, amount=left.amount + right.amount)

    @classmethod
    def generate_pydoc(cls, definitions: List[Tuple[str, str]], inferred_name=None, comparable=False) -> str:
        super(TicketType, cls).generate_pydoc(definitions)
        param_expr = micheline_to_michelson(cls.args[0].as_micheline_expr())
        if cls.args[0].args:
            name = cls.field_name or cls.type_name or inferred_name or f'{cls.prim}_{len(definitions)}'
            param_name = f'{name}_param'
            definitions.insert(0, (param_name, param_expr))
            return f'ticket (${param_name})'
        else:
            return f'ticket ({param_expr})'

    @classmethod
    def dummy(cls, context: AbstractContext):
        raise AssertionError('forbidden')

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'TicketType':
        type_impl = PairType.create_type(args=[AddressType, cls.args[0], NatType])
        comb = type_impl.from_micheline_value(val_expr)
        return cls.from_comb(comb)

    @classmethod
    def from_python_object(cls, py_obj) -> 'MichelsonType':
        type_impl = PairType.create_type(args=[AddressType, cls.args[0], NatType])
        comb = type_impl.from_python_object(py_obj)
        return cls.from_comb(comb)

    def to_comb(self) -> PairType:
        return PairType.from_comb(items=[AddressType(self.ticketer), self.item, NatType(self.amount)])

    def to_literal(self) -> Type[Micheline]:
        return self.to_comb().to_literal()

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return self.to_comb().to_micheline_value(mode=mode)

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return self.ticketer, self.item.to_python_object(try_unpack=try_unpack, comparable=True), self.amount

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'MichelsonType':
        return self

    def split(self, amount_left: int, amount_right: int) -> Optional[Tuple['TicketType', 'TicketType']]:
        if amount_left + amount_right != self.amount:
            return None
        else:
            left = TicketType(ticketer=self.ticketer, item=copy(self.item), amount=amount_left)
            right = TicketType(ticketer=self.ticketer, item=copy(self.item), amount=amount_right)
            return left, right



docs = {   'ABS': 'ABS\nABS :: int : A => nat : A\nObtain the absolute value of an integer',
    'ADD': 'ADD\n'
           'ADD :: nat : nat : A => nat : A\n'
           'ADD :: nat : int : A => int : A\n'
           'ADD :: int : nat : A => int : A\n'
           'ADD :: int : int : A => int : A\n'
           'ADD :: timestamp : int : A => timestamp : A\n'
           'ADD :: int : timestamp : A => timestamp : A\n'
           'ADD :: mutez : mutez : A => mutez : A\n'
           'ADD :: bls12_381_g1 : bls12_381_g1 : A => bls12_381_g1 : A\n'
           'ADD :: bls12_381_g2 : bls12_381_g2 : A => bls12_381_g2 : A\n'
           'ADD :: bls12_381_fr : bls12_381_fr : A => bls12_381_fr : A\n'
           'Add two numerical values',
    'ADDRESS': 'ADDRESS\nADDRESS :: contract ty1 : A => address : A\nPush the address of a contract',
    'AMOUNT': 'AMOUNT\nAMOUNT :: A => mutez : A\nPush the amount of the current transaction',
    'AND': 'AND\n'
           'AND :: bool : bool : A => bool : A\n'
           'AND :: nat : nat : A => nat : A\n'
           'AND :: int : nat : A => nat : A\n'
           'Boolean and bitwise AND',
    'APPLY': 'APPLY\n'
             'APPLY :: ty1 : lambda ( pair ty1 ty2 ) ty3 : A => lambda ty2 ty3 : A\n'
             'Partially apply a tuplified function from the stack',
    'BALANCE': 'BALANCE\nBALANCE :: A => mutez : A\nPush the current amount of mutez of the executing contract',
    'BLAKE2B': 'BLAKE2B\nBLAKE2B :: bytes : A => bytes : A\nCompute a Blake2B cryptographic hash',
    'CAR': 'CAR\nCAR :: pair ty1 ty2 : A => ty1 : A\nAccess the left part of a pair',
    'CAST': '',
    'CDR': 'CDR\nCDR :: pair ty1 ty2 : A => ty2 : A\nAccess the right part of a pair',
    'CHAIN_ID': 'CHAIN_ID\nCHAIN_ID :: A => chain_id : A\nPush the chain identifier',
    'CHECK_SIGNATURE': 'CHECK_SIGNATURE\n'
                       'CHECK_SIGNATURE :: key : signature : bytes : A => bool : A\n'
                       'Verify signature of bytes by key',
    'COMPARE': 'COMPARE\nCOMPARE :: cty : cty : A => int : A\nCompare two values',
    'CONCAT': 'CONCAT\n'
              'CONCAT :: string : string : A => string : A\n'
              'CONCAT :: list string : A => string : A\n'
              'CONCAT :: bytes : bytes : A => bytes : A\n'
              'CONCAT :: list bytes : A => bytes : A\n'
              'Concatenate a string, byte sequence, string list or byte sequence list',
    'CONS': 'CONS\nCONS :: ty1 : list ty1 : A => list ty1 : A\nPrepend an element to a list',
    'CONTRACT': 'CONTRACT ty\n'
                'CONTRACT ty :: address : A => option ( contract ty ) : A\n'
                'Cast an address to a typed contract',
    'CREATE_ACCOUNT': '\nPush an account creation operation',
    'CREATE_CONTRACT': 'CREATE_CONTRACT { parameter ty1 ; storage ty2 ; code instr1 }\n'
                       'CREATE_CONTRACT { parameter ty1 ; storage ty2 ; code instr1 } :: option key_hash : mutez : ty2 '
                       ': A => operation : address : A\n'
                       'Push a contract creation operation',
    'DIG': 'DIG n\nDIG n :: A @ ( ty1 : B ) => ty1 : ( A @ B )\nRetrieve the n\\ th element of the stack',
    'DIP': 'DIP instr\nDIP instr :: ty : B => ty : C\nRun code protecting the top of the stack',
    'DIPN': 'DIP n instr\nDIP n instr :: A @ B => A @ C\nRun code protecting the n topmost elements of the stack',
    'DROP': 'DROP n\nDROP n :: A @ B => B\nDrop the top n elements of the stack',
    'DUG': 'DUG n\nDUG n :: ty1 : ( A @ B ) => A @ ( ty1 : B )\nInsert the top element at depth n',
    'DUP': 'DUP\nDUP :: ty1 : A => ty1 : ty1 : A\nDuplicate the top of the stack',
    'DUPN': 'DUP n\nDUP n :: A @ ty1 : B => ty1 : A @ ty1 : B\nDuplicate the n\\ th element of the stack',
    'EDIV': 'EDIV\n'
            'EDIV :: nat : nat : A => option ( pair nat nat ) : A\n'
            'EDIV :: nat : int : A => option ( pair int nat ) : A\n'
            'EDIV :: int : nat : A => option ( pair int nat ) : A\n'
            'EDIV :: int : int : A => option ( pair int nat ) : A\n'
            'EDIV :: mutez : nat : A => option ( pair mutez mutez ) : A\n'
            'EDIV :: mutez : mutez : A => option ( pair nat mutez ) : A\n'
            'Euclidean division',
    'EMPTY_BIG_MAP': 'EMPTY_BIG_MAP kty vty\n'
                     'EMPTY_BIG_MAP kty vty :: A => big_map kty vty : A\n'
                     'Build a new, empty big_map from kty to vty',
    'EMPTY_MAP': 'EMPTY_MAP kty vty\nEMPTY_MAP kty vty :: A => map kty vty : A\nBuild a new, empty map from kty to vty',
    'EMPTY_SET': 'EMPTY_SET cty\nEMPTY_SET cty :: A => set cty : A\nBuild a new, empty set for elements of type cty',
    'EQ': 'EQ\nEQ :: int : A => bool : A\nCheck that the top of the stack equals zero',
    'EXEC': 'EXEC\nEXEC :: ty1 : lambda ty1 ty2 : A => ty2 : A\nExecute a function from the stack',
    'EXPAND': '',
    'FAILWITH': 'FAILWITH\nFAILWITH :: ty1 : A => B\nExplicitly abort the current program',
    'GE': 'GE\nGE :: int : A => bool : A\nCheck that the top of the stack is greater than or equal to zero',
    'GET': 'GET\n'
           'GET :: kty : map kty vty : A => option vty : A\n'
           'GET :: kty : big_map kty vty : A => option vty : A\n'
           'Access an element in a map or big_map',
    'GETN': 'GET n\n'
            'GET ( 0 ) :: ty : A => ty : A\n'
            "GET ( 2 * n ) :: pair ty0 .. tyN ty' : A => ty' : A\n"
            "GET ( 2 * n + 1 ) :: pair ty0 .. tyN ty' ty'' : A => ty' : A\n"
            'Access an element or a sub comb in a right comb',
    'GET_AND_UPDATE': 'GET_AND_UPDATE\n'
                      'GET_AND_UPDATE :: kty : option vty : map kty vty : A => option vty : map kty vty : A\n'
                      'GET_AND_UPDATE :: kty : option vty : big_map kty vty : A => option vty : big_map kty vty : A\n'
                      'A combination of the GET and UPDATE instructions',
    'GT': 'GT\nGT :: int : A => bool : A\nCheck that the top of the stack is greater than zero',
    'HASH_KEY': 'HASH_KEY\nHASH_KEY :: key : A => key_hash : A\nCompute the Base58Check of a public key',
    'IF': 'IF instr1 instr2\nIF instr1 instr2 :: bool : A => B\nConditional branching',
    'IF_CONS': 'IF_CONS instr1 instr2\nIF_CONS instr1 instr2 :: list ty : A => B\nInspect a list',
    'IF_LEFT': 'IF_LEFT instr1 instr2\nIF_LEFT instr1 instr2 :: or ty1 ty2 : A => B\nInspect a value of a union',
    'IF_NONE': 'IF_NONE instr1 instr2\nIF_NONE instr1 instr2 :: option ty1 : A => B\nInspect an optional value',
    'IMPLICIT_ACCOUNT': 'IMPLICIT_ACCOUNT\n'
                        'IMPLICIT_ACCOUNT :: key_hash : A => contract unit : A\n'
                        'Create an implicit account',
    'INT': 'INT\n'
           'INT :: nat : A => int : A\n'
           'INT :: bls12_381_fr : A => int : A\n'
           'Convert a natural number or a BLS12-381 field element to an integer',
    'ISNAT': 'ISNAT\nISNAT :: int : A => option nat : A\nConvert a non-negative integer to a natural number',
    'ITER': 'ITER instr\n'
            'ITER instr :: list ty : A => A\n'
            'ITER instr :: set cty : A => A\n'
            'ITER instr :: map kty vty : A => A\n'
            'Iterate over a set, list or map',
    'JOIN_TICKETS': 'JOIN_TICKETS\n'
                    'JOIN_TICKETS :: pair ( ticket cty ) ( ticket cty ) : A => option ( ticket cty ) : A\n'
                    'Join two tickets into one',
    'KECCAK': 'KECCAK\nKECCAK :: bytes : A => bytes : A\nCompute a Keccak-256 cryptographic hash',
    'LAMBDA': 'LAMBDA ty1 ty2 instr\nLAMBDA ty1 ty2 instr :: A => lambda ty1 ty2 : A\nPush a lambda onto the stack',
    'LE': 'LE\nLE :: int : A => bool : A\nCheck that the top of the stack is less than or equal to zero',
    'LEFT': 'LEFT ty2\nLEFT ty2 :: ty1 : A => or ty1 ty2 : A\nWrap a value in a union (left case)',
    'LEVEL': 'LEVEL\nLEVEL :: A => nat : A\nPush the current block level',
    'LOOP': 'LOOP instr\nLOOP instr :: bool : A => A\nA generic loop',
    'LOOP_LEFT': 'LOOP_LEFT instr\nLOOP_LEFT instr :: or ty1 ty2 : A => ty2 : A\nLoop with accumulator',
    'LSL': 'LSL\nLSL :: nat : nat : A => nat : A\nLogically left shift a natural number',
    'LSR': 'LSR\nLSR :: nat : nat : A => nat : A\nLogically right shift a natural number',
    'LT': 'LT\nLT :: int : A => bool : A\nCheck that the top of the stack is less than zero',
    'MAP': 'MAP instr\n'
           'MAP instr :: list ty : A => list ty2 : A\n'
           'MAP instr :: map kty ty1 : A => map kty ty2 : A\n'
           'Apply instr to each element of a list or map.',
    'MEM': 'MEM\n'
           'MEM :: cty : set cty : A => bool : A\n'
           'MEM :: kty : map kty vty : A => bool : A\n'
           'MEM :: kty : big_map kty vty : A => bool : A\n'
           'Check for the presence of a binding for a key in a map, set or big_map',
    'MUL': 'MUL\n'
           'MUL :: nat : nat : A => nat : A\n'
           'MUL :: nat : int : A => int : A\n'
           'MUL :: int : nat : A => int : A\n'
           'MUL :: int : int : A => int : A\n'
           'MUL :: mutez : nat : A => mutez : A\n'
           'MUL :: nat : mutez : A => mutez : A\n'
           'MUL :: bls12_381_g1 : bls12_381_fr : A => bls12_381_g1 : A\n'
           'MUL :: bls12_381_g2 : bls12_381_fr : A => bls12_381_g2 : A\n'
           'MUL :: bls12_381_fr : bls12_381_fr : A => bls12_381_fr : A\n'
           'MUL :: nat : bls12_381_fr : A => bls12_381_fr : A\n'
           'MUL :: int : bls12_381_fr : A => bls12_381_fr : A\n'
           'MUL :: bls12_381_fr : nat : A => bls12_381_fr : A\n'
           'MUL :: bls12_381_fr : int : A => bls12_381_fr : A\n'
           'Multiply two numerical values',
    'NEG': 'NEG\n'
           'NEG :: nat : A => int : A\n'
           'NEG :: int : A => int : A\n'
           'NEG :: bls12_381_g1 : A => bls12_381_g1 : A\n'
           'NEG :: bls12_381_g2 : A => bls12_381_g2 : A\n'
           'NEG :: bls12_381_fr : A => bls12_381_fr : A\n'
           'Negate a numerical value',
    'NEQ': 'NEQ\nNEQ :: int : A => bool : A\nCheck that the top of the stack is not equal to zero',
    'NEVER': 'NEVER\nNEVER :: never : A => B\nClose a forbidden branch',
    'NIL': 'NIL ty\nNIL ty :: A => list ty : A\nPush an empty list',
    'NONE': 'NONE ty\nNONE ty :: A => option ty : A\nPush the absent optional value',
    'NOOP': '{}\n{} :: A => A\nEmpty instruction sequence',
    'NOT': 'NOT\n'
           'NOT :: bool : A => bool : A\n'
           'NOT :: nat : A => int : A\n'
           'NOT :: int : A => int : A\n'
           'Boolean negation and bitwise complement',
    'NOW': 'NOW\nNOW :: A => timestamp : A\nPush block timestamp',
    'OR': 'OR\nOR :: bool : bool : A => bool : A\nOR :: nat : nat : A => nat : A\nBoolean and bitwise OR',
    'PACK': 'PACK\nPACK :: ty : A => bytes : A\nSerialize data',
    'PAIR': "PAIR\nPAIR :: ty1 : ty2 : A => pair ty1 ty2 : A\nBuild a pair from the stack's top two elements",
    'PAIRING_CHECK': 'PAIRING_CHECK\n'
                     'PAIRING_CHECK :: list ( pair bls12_381_g1 bls12_381_g2 ) : A => bool : A\n'
                     'Check a BLS12-381 pairing',
    'PAIRN': 'PAIR n\n'
             'PAIR n :: ty1 : .... : tyN : A => pair ty1 .... tyN : A\n'
             'Fold n values on the top of the stack into a right comb',
    'PUSH': 'PUSH ty x\nPUSH ty x :: A => ty1 : A\nPush a constant value of a given type onto the stack',
    'READ_TICKET': 'READ_TICKET\n'
                   'READ_TICKET :: ticket cty : A => pair address cty nat : ticket cty : A\n'
                   'Retrieve the information stored in a ticket. Also return the ticket.',
    'RENAME': '',
    'RIGHT': 'RIGHT ty1\nRIGHT ty1 :: ty2 : A => or ty1 ty2 : A\nWrap a value in a union (right case)',
    'SAPLING_EMPTY_STATE': 'SAPLING_EMPTY_STATE ms\n'
                           'SAPLING_EMPTY_STATE ms :: A => sapling_state ms : A\n'
                           'Pushes an empty Sapling state on the stack',
    'SAPLING_VERIFY_UPDATE': 'SAPLING_VERIFY_UPDATE\n'
                             'SAPLING_VERIFY_UPDATE :: sapling_transaction ms : sapling_state ms : A => option ( pair '
                             'int ( sapling_state ms ) ) : A\n'
                             'Verify and apply a transaction on a Sapling state',
    'SELF': 'SELF\nSELF :: A => contract ty : A\nPush the current contract',
    'SELF_ADDRESS': 'SELF_ADDRESS\nSELF_ADDRESS :: A => address : A\nPush the address of the current contract',
    'SENDER': 'SENDER\nSENDER :: A => address : A\nPush the contract that initiated the current internal transaction',
    'SEQ': 'instr1 ; instr2\ninstr1 ; instr2 :: A => C\nInstruction sequence',
    'SET_DELEGATE': 'SET_DELEGATE\nSET_DELEGATE :: option key_hash : A => operation : A\nPush a delegation operation',
    'SHA256': 'SHA256\nSHA256 :: bytes : A => bytes : A\nCompute a SHA-256 cryptographic hash',
    'SHA3': 'SHA3\nSHA3 :: bytes : A => bytes : A\nCompute a SHA3-256 cryptographic hash',
    'SHA512': 'SHA512\nSHA512 :: bytes : A => bytes : A\nCompute a SHA-512 cryptographic hash',
    'SIZE': 'SIZE\n'
            'SIZE :: set cty : A => nat : A\n'
            'SIZE :: map kty vty : A => nat : A\n'
            'SIZE :: list ty : A => nat : A\n'
            'SIZE :: string : A => nat : A\n'
            'SIZE :: bytes : A => nat : A\n'
            'Obtain size of a string, list, set, map or byte sequence bytes',
    'SLICE': 'SLICE\n'
             'SLICE :: nat : nat : string : A => option string : A\n'
             'SLICE :: nat : nat : bytes : A => option bytes : A\n'
             'Obtain a substring or subsequence of a string respectively byte sequence bytes',
    'SOME': 'SOME\nSOME :: ty1 : A => option ty1 : A\nWrap an existing optional value',
    'SOURCE': 'SOURCE\nSOURCE :: A => address : A\nPush the contract that initiated the current transaction',
    'SPLIT_TICKET': 'SPLIT_TICKET\n'
                    'SPLIT_TICKET :: ticket cty : pair nat nat : A => option ( pair ( ticket cty ) ( ticket cty ) ) : '
                    'A\n'
                    'Split a ticket in two',
    'STEPS_TO_QUOTA': '\nPush the remaining steps before the contract execution must terminate',
    'SUB': 'SUB\n'
           'SUB :: nat : nat : A => int : A\n'
           'SUB :: nat : int : A => int : A\n'
           'SUB :: int : nat : A => int : A\n'
           'SUB :: int : int : A => int : A\n'
           'SUB :: timestamp : int : A => timestamp : A\n'
           'SUB :: timestamp : timestamp : A => int : A\n'
           'SUB :: mutez : mutez : A => mutez : A\n'
           'Subtract two numerical values',
    'SWAP': 'SWAP\nSWAP :: ty1 : ty2 : A => ty2 : ty1 : A\nSwap the top two elements of the stack',
    'TICKET': 'TICKET\nTICKET :: cty : nat : A => ticket cty : A\nCreate a ticket',
    'TOP': '',
    'TOTAL_VOTING_POWER': 'TOTAL_VOTING_POWER\n'
                          'TOTAL_VOTING_POWER :: A => nat : A\n'
                          'Return the total voting power of all contracts',
    'TRANSFER_TOKENS': 'TRANSFER_TOKENS\n'
                       'TRANSFER_TOKENS :: ty : mutez : contract ty : A => operation : A\n'
                       'Push a transaction operation',
    'UNIT': 'UNIT\nUNIT :: A => unit : A\nPush the unit value onto the stack',
    'UNPACK': 'UNPACK ty\nUNPACK ty :: bytes : A => option ty : A\nDeserialize data, if valid',
    'UNPAIR': 'UNPAIR\nUNPAIR :: pair ty1 ty2 : A => ty1 : ty2 : A\nSplit a pair into its components',
    'UNPAIRN': 'UNPAIR n\n'
               'UNPAIR n :: pair ty1 .... tyN : A => ty1 : .... : tyN : A\n'
               'Unfold n values from a right comb on the top of the stack',
    'UPDATE': 'UPDATE\n'
              'UPDATE :: cty : bool : set cty : A => set cty : A\n'
              'UPDATE :: kty : option vty : map kty vty : A => map kty vty : A\n'
              'UPDATE :: kty : option vty : big_map kty vty : A => big_map kty vty : A\n'
              'Add or remove an element in a map, big_map or set',
    'UPDATEN': 'UPDATE n\n'
               'UPDATE 0 :: ty1 : ty2 : A => ty1 : A\n'
               "UPDATE ( 2 * n ) :: ty' : pair ty0 .. tyN ty'' : A => pair ty0 .. tyN ty' : A\n"
               "UPDATE ( 2 * n + 1 ) :: ty' : pair ty0 .. tyN ty'' ty''' : A => pair ty0 .. tyN ty' ty''' : A\n"
               'Update an element or a sub comb in a right comb',
    'VOTING_POWER': 'VOTING_POWER\n'
                    'VOTING_POWER :: key_hash : A => nat : A\n'
                    'Return the voting power of a given contract',
    'XOR': 'XOR\nXOR :: bool : bool : A => bool : A\nXOR :: nat : nat : A => nat : A\nBoolean and bitwise eXclusive OR',
    'address': 'address\nAddress of an untyped contract',
    'big_map': 'big_map kty vty\nA lazily deserialized map from kty to vty',
    'bls12_381_fr': 'bls12_381_fr\nAn element of the BLS12-381 scalar field F\\ :sub:r',
    'bls12_381_g1': 'bls12_381_g1\nA point on the BLS12-381 curve G\\ :sub:1',
    'bls12_381_g2': 'bls12_381_g2\nA point on the BLS12-381 curve G\\ :sub:2',
    'bool': 'bool\nA boolean',
    'bytes': 'bytes\nA sequence of bytes',
    'chain_id': 'chain_id\nA chain identifier',
    'contract': "contract type\nAddress of a contract, where type is the contract's parameter type",
    'int': 'int\nAn arbitrary-precision integer',
    'key': 'key\nA public cryptographic key',
    'key_hash': 'key_hash\nA hash of a public cryptographic key',
    'lambda': 'lambda ty1 ty2\nA lambda with given parameter and return types',
    'list': 'list type\nA single, immutable, homogeneous linked list',
    'map': 'map kty vty\nAn immutable map from kty to vty',
    'mutez': 'mutez\nA specific type for manipulating tokens',
    'nat': 'nat\nAn arbitrary-precision natural number',
    'never': 'never\nThe empty type',
    'operation': 'operation\nAn internal operation emitted by a contract',
    'option': 'option ty\nAn optional value',
    'or': 'or ty1 ty2\nA union of two types',
    'pair': 'pair ty1 ty2\nA pair or tuple of values',
    'sapling_state': 'sapling_state n\nA Sapling state',
    'sapling_transaction': 'sapling_transaction n\nA Sapling transaction',
    'set': 'set cty\nAn immutable set of comparable values of type cty',
    'signature': 'signature\nA cryptographic signature',
    'string': 'string\nA string of characters',
    'ticket': 'ticket cty\nA ticket used to authenticate information of type cty',
    'timestamp': 'timestamp\nA real-world date.',
    'unit': 'unit\nThe type whose only value is Unit'}



from binascii import hexlify
from datetime import datetime
from functools import cached_property
from functools import lru_cache
from time import sleep
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional

import requests
import simplejson as json
from deprecation import deprecated  # type: ignore

from pytezos.crypto.encoding import base58_decode
from pytezos.jupyter import get_attr_docstring
from pytezos.logging import logger
from pytezos.rpc.kind import validation_passes
from pytezos.rpc.protocol import BlockQuery
from pytezos.rpc.protocol import BlocksQuery
from pytezos.rpc.query import RpcQuery
from pytezos.rpc.search import CyclesQuery
from pytezos.rpc.search import VotingPeriodsQuery

MAX_BLOCK_TIMEOUT = 86400


def make_operation_result(**kwargs):
    return {'metadata': {'operation_result': kwargs}}


class ShellQuery(RpcQuery, path=''):
    @property
    def blocks(self) -> BlocksQuery:
        """Shortcut for `chains.main.blocks`"""
        return self.chains.main.blocks

    @property
    def head(self) -> BlockQuery:
        """Shortcut for `blocks.head`"""
        return self.blocks.head

    @cached_property
    def block(self) -> BlockQuery:
        """Cached head block, useful if you just want to explore things."""
        return self.blocks[self.head.hash()]

    @property
    def cycles(self):
        """Operate on cycles rather than blocks."""
        return CyclesQuery(
            node=self.node,
            path=self._wild_path + '/chains/{}/blocks',
            params=self._params + ['main'],
        )

    @property
    def voting_periods(self):
        """
        Operate on voting periods rather than blocks.
        """
        return VotingPeriodsQuery(
            node=self.node,
            path=self._wild_path + '/chains/{}/blocks',
            params=self._params + ['main'],
        )

    @property
    def contracts(self):
        """Shortcut for `head.context.contracts`"""
        return self.head.context.contracts

    @property
    def mempool(self):
        """Shortcut for `chains.main.mempool`"""
        return self.chains.main.mempool

    def wait_blocks(
        self,
        current_block_hash: str,
        max_blocks: int = 1,
        yield_current=False,
        time_between_blocks: Optional[int] = None,
        block_timeout: Optional[int] = None,
    ) -> Generator[str, None, None]:
        """Iterates over future blocks (waits and yields block hash), handles reorgs

        :param current_block_hash: hash of the current block (head)
        :param max_blocks: number of blocks to iterate (not including the current one)
        :param yield_current: yield current block hash at the very beginning
        :param time_between_blocks: override protocol constant
        :param block_timeout: set block timeout (by default Pytezos will wait for a long time)
        :return: block hashes
        """
        if time_between_blocks is None:
            constants = self.blocks[current_block_hash].context.constants()
            time_between_blocks = int(constants.get('minimal_block_delay', 0))

        if block_timeout is None:
            block_timeout = MAX_BLOCK_TIMEOUT

        if yield_current:
            yield current_block_hash

        current_header = self.blocks[current_block_hash].header()
        max_level = current_header['level'] + max_blocks

        while current_header['level'] < max_level:
            logger.info('Current level: %d (max %d)', current_header['level'], max_level)
            prev_block_dt = datetime.strptime(current_header['timestamp'], '%Y-%m-%dT%H:%M:%SZ')
            elapsed_sec = (datetime.utcnow() - prev_block_dt).seconds
            sleep_sec = 1 if elapsed_sec > time_between_blocks else (time_between_blocks - elapsed_sec + 1)

            logger.info('Sleep %d seconds until block %s is superseded', sleep_sec, current_block_hash)
            sleep(sleep_sec)

            next_block_hash: Optional[str] = None

            for delay in range(block_timeout):
                next_block_hash = self.head.hash()
                if current_block_hash == next_block_hash:
                    sleep(1)
                else:
                    logger.info('Found new block %s (%d sec delay)', next_block_hash, delay)
                    break

            if current_block_hash != next_block_hash:
                assert next_block_hash
                yield next_block_hash
                current_block_hash = next_block_hash
                current_header = self.blocks[current_block_hash].header()
            else:
                raise TimeoutError('Reached timeout (%d sec) while waiting for the next block', block_timeout)

    def wait_operations(
        self,
        opg_hashes: List[str],
        ttl: int,
        min_confirmations: int,
        current_block_hash: Optional[str] = None,
        time_between_blocks: Optional[int] = None,
        block_timeout: Optional[int] = None,
    ) -> List[dict]:
        """Wait for one or many operations gain enough confirmations

        :param opg_hashes: list of operation hashes
        :param ttl: max time-to-live value (in mempool)
        :param min_confirmations: minimum number of blocks after inclusion to wait for
        :param current_block_hash: current block hash (head)
        :param time_between_blocks: override protocol constant
        :param block_timeout: set block timeout (by default Pytezos will wait for a long time)
        :return: list of operation contents with metadata
        """

        confirmations = {}
        pending = set(opg_hashes)
        operations = []
        block_hash = current_block_hash

        if block_hash is None:
            block_hash = self.head.hash()

        for block_hash in self.wait_blocks(  # noqa: B020
            current_block_hash=block_hash,
            max_blocks=ttl,
            yield_current=True,
            time_between_blocks=time_between_blocks,
            block_timeout=block_timeout,
        ):
            if len(pending) > 0:
                mempool = set(map(lambda x: x['hash'], self.mempool.pending_operations.flatten()))
                for opg_hash in opg_hashes:
                    if opg_hash in pending:
                        if opg_hash in mempool:
                            logger.info('Operation %s is still in mempool', opg_hash)
                        else:
                            logger.info('Operation %s has left mempool', opg_hash)
                            pending.remove(opg_hash)

            included = self.blocks[block_hash].operation_hashes()
            for i, vp in enumerate(included):
                for j, opg_hash in enumerate(vp):
                    if opg_hash in opg_hashes and opg_hash not in confirmations:
                        logger.info('Operation %s has been included to block %s', opg_hash, block_hash)
                        if opg_hash in pending:  # can be still in the particular node's mempool (not yet removed)
                            pending.remove(opg_hash)
                        confirmations[opg_hash] = 0  # initialize
                        operations.append(self.blocks[block_hash].operations[i][j]())

            for opg_hash in confirmations:
                confirmations[opg_hash] += 1
                logger.info(
                    'Operation %s has %d/%d confirmations', opg_hash, confirmations[opg_hash], min_confirmations
                )

            if len(operations) == len(opg_hashes):
                break

        if len(operations) < len(opg_hashes):
            raise StopIteration('Only %d of %d operations were included, stopping', len(operations), len(opg_hashes))

        for _ in self.wait_blocks(
            block_hash,
            max_blocks=min_confirmations - 1,
            time_between_blocks=time_between_blocks,
        ):
            for opg_hash in opg_hashes:
                confirmations[opg_hash] += 1
                logger.info(
                    'Operation %s has %d/%d confirmations', opg_hash, confirmations[opg_hash], min_confirmations
                )

        return operations

    @deprecated(deprecated_in='3.2.2', removed_in='4.0.0', details=f'Use wait_blocks() instead')
    def wait_next_block(
        self,
        delay_sec=1,
        prev_hash=None,
        time_between_blocks: Optional[int] = None,
        max_iterations: Optional[int] = None,
    ):
        """Wait until next block is finalized.

        :param prev_hash: Current block hash (optional). If not set, current head is used.
        :param time_between_blocks: override the corresponding parameter from constants
        :param max_iterations: Manually set the number of iterations
        :param delay_sec: Sleep delay
        """
        if time_between_blocks is None:
            time_between_blocks = int(self.block.context.constants()["minimal_block_delay"])  # type: ignore

        if time_between_blocks > 0:
            if prev_hash is None:
                header = self.head.header()
                prev_hash = header['hash']
                prev_block_dt = datetime.strptime(header['timestamp'], '%Y-%m-%dT%H:%M:%SZ')
                elapsed_sec = (datetime.utcnow() - prev_block_dt).seconds
                sleep_sec = 0 if elapsed_sec > time_between_blocks else time_between_blocks - elapsed_sec
            else:
                sleep_sec = time_between_blocks
            logger.info('Wait %s seconds until block %s is finalized', sleep_sec, prev_hash)
            sleep(sleep_sec)

        if max_iterations is None:
            max_iterations = max(1, time_between_blocks)

        for _ in range(max_iterations):
            current_block_hash = self.head.hash()
            if current_block_hash == prev_hash:
                sleep(delay_sec)
            else:
                return current_block_hash
        raise StopIteration("Timeout")

    def get_confirmations(self, opg_hash, kind, branch, head) -> int:
        """Returns the number of blocks applied after the operation was included in chain

        :param opg_hash: Operation group hash
        :param kind: Operation kind ('transaction', 'origination', etc)
        :param branch: Block ID one should stop the search at
        :param head: Block ID one should start the search from
        :return: Number of confirmations (0 if not found)
        """
        start = self.blocks[head].header()['level']
        stop = self.blocks[branch].header()['level']
        for level in range(start, stop, -1):
            vp = validation_passes[kind]
            hashes = self.blocks[level].operation_hashes[vp]()
            for idx, _hash in enumerate(hashes):
                if opg_hash == _hash:
                    _ = self.blocks[level].operations[vp, idx]()
                    logger.info('Operation %s was included in block %s', opg_hash, level)
                    return start - level + 1
        return 0


class ChainQuery(RpcQuery, path='/chains/{}'):
    def watermark(self):
        """Chain watermark, hex encoded."""
        data = self.chain_id()
        return hexlify(base58_decode(data.encode())).decode()


class InvalidBlockQuery(RpcQuery, path='/chains/{}/invalid_blocks/{}'):
    def delete(self):
        return self._delete()


class MempoolQuery(RpcQuery, path='/chains/{}/mempool'):
    def post(self, configuration):
        """Set operation filter rules.

        :param configuration: a JSON dictionary, known keys are `minimal_fees`, `minimal_nanotez_per_gas_unit`,
            `minimal_nanotez_per_byte`
        """
        return self._post(json=configuration)


class PendingOperationsQuery(RpcQuery, path='/chains/{}/mempool/pending_operations'):
    def __getitem__(self, item: str) -> Dict[str, Any]:
        """Search for operation in node's mempool by hash.

        :param item: operation group hash (base58)
        """
        operations_dict = self()
        for status, operations in operations_dict.items():
            for operation in operations:
                if isinstance(operation, dict):
                    if operation['hash'] == item:
                        return {
                            **make_operation_result(status=status),
                            **operation,
                        }
                elif isinstance(operation, list):
                    if operation[0] == item:
                        errors = operation[1].pop1('error', default=[])
                        return {
                            **make_operation_result(status=status, errors=errors),
                            **operation[1],
                            'hash': operation[0],
                        }
                else:
                    raise Exception('Unknown operation type', operation)
        raise StopIteration

    def flatten(self) -> List[Dict[str, Any]]:
        operations_dict = self()
        operations_list = []
        for status, operations in operations_dict.items():
            for operation in operations:
                if isinstance(operation, dict):
                    operations_list.append(
                        {
                            **make_operation_result(status=status),
                            **operation,
                        }
                    )
                elif isinstance(operation, list):
                    errors = operation[1].pop('error', [])
                    operations_list.append(
                        {
                            **make_operation_result(status=status, errors=errors),
                            **operation[1],
                            'hash': operation[0],
                        }
                    )
                else:
                    raise Exception('Unknown operation type', operation)
        return operations_list

    def __repr__(self):
        res = [
            super(PendingOperationsQuery, self).__repr__(),
            '[]' + get_attr_docstring(self.__class__, '__getitem__'),
        ]
        return '\n'.join(res)


class DescribeQuery(RpcQuery, path='/describe'):
    def __call__(self, recurse=True):
        """Get RPCs documentation and input/output schema.

        :param recurse: Show information for child elements, default is True.
            In some cases doesn't work without this flag.
        """
        return super().__call__(recurse=recurse)

    def __repr__(self):
        res = [
            super().__repr__(),
            f'(){get_attr_docstring(DescribeQuery, "__call__")}',
            'Can be followed by any path:\n.chains\n.network.connections\netc\n',
        ]
        return '\n'.join(res)


class BlockInjectionQuery(RpcQuery, path='/injection/block'):
    def post(self, block, timestamp=None, _async=False, force=False, chain=None):
        """Inject a block in the node and broadcast it.

        The `operations` embedded in `blockHeader` might be pre-validated using a contextual RPCs from the latest block
        (e.g. '/blocks/head/context/preapply').

        block format:

        .. code-block:: python

            {
                "data": <hex-encoded block header>,
                "operations": [ [ {
                "branch": <block_hash>,
                "data": <hex-encoded operation>
                } ... ] ... ]
            }

        :param block: JSON input
        :param _async: By default, the RPC will wait for the block to be validated before answering, \
        set True if you don't want to.
        :param force:
        :param chain: Optionally you can specify the chain (main/test)
        :returns: ID of the block
        """
        return self._post(
            params={
                'async': _async,
                'force': force,
                'chain': chain,
                'timestamp': timestamp,
            },
            json=block,
        )


class OperationInjectionQuery(RpcQuery, path='/injection/operation'):
    def post(self, operation, _async=False, chain=None):
        """Inject an operation in node and broadcast it.
        The `signedOperationContents` should be constructed using a contextual RPCs from the latest block
        and signed by the client.

        :param operation: Hex-encoded operation data or bytes
        :param _async: By default, the RPC will wait for the operation to be (pre-)validated before answering,
            set True if you don't want to.
        :param chain: Optionally you can specify the chain
        :returns: ID of the operation
        """
        if isinstance(operation, bytes):
            operation = operation.hex()

        return self._post(
            params={
                'async': _async,
                'chain': chain,
            },
            json=operation,
        )


class ProtocolInjectionQuery(RpcQuery, path='/injection/protocol'):
    def post(self, protocol, _async=False, force=False):
        """Inject a protocol in node.

        protocol format:

        .. code-block:: python

            {
                "expected_env_version": <integer>,
                "components": [{
                    "name": <unistring>,
                    "interface"?: <hex-encoded data>,
                    "implementation": <hex-encoded data> }
                     ...
                ]}
            }

        :param protocol: JSON input
        :param _async:
        :param force:
        :returns: ID of the protocol
        """
        return self._post(
            params={
                'async': _async,
                'force': force,
            },
            json=protocol,
        )


class ResponseGenerator:
    def __init__(self, res: requests.Response):
        self._lines = res.iter_lines()

    def __iter__(self):
        for line in self._lines:
            yield json.loads(line.decode())


class MonitorQuery(
    RpcQuery,
    path=[
        '/monitor/active_chains',
        '/monitor/bootstrapped',
        '/monitor/commit_hash',
        '/monitor/heads/{}',
        '/monitor/protocols',
        '/monitor/valid_blocks',
    ],
):
    def __call__(self, *args, **kwargs):
        return ResponseGenerator(
            self.node.request(
                method='GET',
                path=self.path,
                params=kwargs,
                stream=True,
            )
        )

    def __repr__(self):
        res = [
            super().__repr__(),
            'NOTE: Returned object is a generator.',
        ]
        return '\n'.join(res)


class ConnectionQuery(RpcQuery, path='/network/connections/{}'):
    def delete(self, wait=False):
        return self._delete(params={'wait': wait})


class NetworkItems(RpcQuery, path=['/network/peers', '/network/points']):
    def __call__(self, _filter=None):
        return self._get(params={'filter': _filter})


class NetworkLogQuery(RpcQuery, path=['/network/peers/{}/log', '/network/points/{}/log']):
    def __call__(self, monitor=False):
        if monitor:
            return ResponseGenerator(
                self.node.request(
                    method='GET',
                    path=self.path,
                    stream=True,
                )
            )
        return self._get()



from typing import List
from typing import Tuple
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import BigMapType
from pytezos.michelson.types import BoolType
from pytezos.michelson.types import ListType
from pytezos.michelson.types import MapType
from pytezos.michelson.types import MichelsonType
from pytezos.michelson.types import OptionType
from pytezos.michelson.types import SetType


class ConsInstruction(MichelsonInstruction, prim='CONS'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        elt, lst = cast(Tuple[MichelsonType, ListType], stack.pop2())
        lst.assert_type_in(ListType)
        res = lst.prepend(elt)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [elt, lst], [res]))  # type: ignore
        return cls(stack_items_added=1)


class NilInstruction(MichelsonInstruction, prim='NIL', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = ListType.empty(cls.args[0])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class EmptyBigMapInstruction(MichelsonInstruction, prim='EMPTY_BIG_MAP', args_len=2):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = BigMapType.empty(key_type=cls.args[0], val_type=cls.args[1])  # type: ignore
        res.attach_context(context)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class EmptyMapInstruction(MichelsonInstruction, prim='EMPTY_MAP', args_len=2):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = MapType.empty(key_type=cls.args[0], val_type=cls.args[1])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class EmptySetInstruction(MichelsonInstruction, prim='EMPTY_SET', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = SetType.empty(item_type=cls.args[0])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class GetInstruction(MichelsonInstruction, prim='GET'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        key, src = cast(Tuple[MichelsonType, Union[MapType, BigMapType]], stack.pop2())
        src.assert_type_in(MapType, BigMapType)
        val = src.get(key, dup=True)
        if val is None:
            res = OptionType.none(src.args[1])
        else:
            res = OptionType.from_some(val)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [key, src], [res]))  # type: ignore
        return cls(stack_items_added=1)


class GetAndUpdateInstruction(MichelsonInstruction, prim='GET_AND_UPDATE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        key, val, src = cast(Tuple[MichelsonType, OptionType, Union[MapType, BigMapType]], stack.pop3())
        src.assert_type_in(MapType, BigMapType)
        prev_val, dst = src.update(key, None if val.is_none() else val.get_some())
        res = OptionType.none(src.args[1]) if prev_val is None else OptionType.from_some(prev_val)
        stack.push(dst)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [key, val, src], [res, dst]))  # type: ignore
        return cls(stack_items_added=2)


class UpdateInstruction(MichelsonInstruction, prim='UPDATE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        key, val, src = cast(
            Tuple[
                MichelsonType,
                Union[OptionType, BoolType],
                Union[MapType, BigMapType, SetType],
            ],
            stack.pop3(),
        )
        val.assert_type_in(OptionType, BoolType)
        if isinstance(val, BoolType):
            src.assert_type_in(SetType)
            dst = src.add(key) if bool(val) else src.remove(key)  # type: ignore
        else:
            src.assert_type_in(MapType, BigMapType)
            _, dst = src.update(key, None if val.is_none() else val.get_some())  # type: ignore
        stack.push(dst)
        stdout.append(format_stdout(cls.prim, [key, val, src], [dst]))  # type: ignore
        return cls(stack_items_added=1)


class MemInstruction(MichelsonInstruction, prim='MEM'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        key, src = cast(Tuple[MichelsonType, Union[SetType, MapType, BigMapType]], stack.pop2())
        src.assert_type_in(MapType, BigMapType, SetType)
        res = BoolType.from_value(src.contains(key))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [key, src], [res]))  # type: ignore
        return cls(stack_items_added=1)


class NoneInstruction(MichelsonInstruction, prim='NONE', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = OptionType.none(cls.args[0])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SomeInstruction(MichelsonInstruction, prim='SOME'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        some = stack.pop1()
        res = OptionType.from_some(some)
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [some], [res]))  # type: ignore
        return cls(stack_items_added=1)



# NOTE: Explaination: https://pytezos.baking-bad.org/tutorials/02.html#operation-group
validation_passes = {
    'failing_noop': -1,
    'endorsement': 0,
    'endorsement_with_slot': 0,
    'proposals': 1,
    'ballot': 1,
    'seed_nonce_revelation': 2,
    'double_endorsement_evidence': 2,
    'double_baking_evidence': 2,
    'activate_account': 2,
    'reveal': 3,
    'transaction': 3,
    'origination': 3,
    'delegation': 3,
    'register_global_constant': 3,
}
operation_tags = {
    'endorsement': 0,
    'endorsement_with_slot': 10,
    'proposals': 5,
    'ballot': 6,
    'seed_nonce_revelation': 1,
    'double_endorsement_evidence': 2,
    'double_baking_evidence': 3,
    'activate_account': 4,
    'failing_noop': 17,
    'reveal': 107,
    'transaction': 108,
    'origination': 109,
    'delegation': 110,
    'register_global_constant': 111,
}



from copy import copy
from copy import deepcopy
from typing import Callable
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.forge import forge_script_expr
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import parse_micheline_literal
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import Undefined
from pytezos.michelson.types.map import EltLiteral
from pytezos.michelson.types.map import MapType


class BigMapType(MapType, prim='big_map', args_len=2):
    def __init__(
        self,
        items: List[Tuple[MichelsonType, MichelsonType]],
        ptr: Optional[int] = None,
        removed_keys: Optional[List[MichelsonType]] = None,
    ):
        super(BigMapType, self).__init__(items=items)
        self.ptr = ptr
        self.removed_keys = removed_keys or []
        self.context: Optional[AbstractContext] = None

    def __len__(self):
        return len(self.items) + len(self.removed_keys)

    def __iter__(self) -> Generator[Tuple[MichelsonType, Optional[MichelsonType]], None, None]:  # type: ignore
        yield from iter(self.items)
        for key in self.removed_keys:
            yield key, None

    def __repr__(self):
        if self.context:
            return f'<{self.ptr}>'
        else:
            elements = [f'{repr(k)}: {repr(v)}' for k, v in self]
            return f'{{{", ".join(elements)}}}'

    def __deepcopy__(self, memodict):
        return self.duplicate()

    def __getitem__(self, key_obj) -> Optional[MichelsonType]:  # type: ignore
        key = self.args[0].from_python_object(key_obj)
        return self.get(key, dup=False)

    @staticmethod
    def empty(key_type: Type[MichelsonType], val_type: Type[MichelsonType]) -> 'BigMapType':
        cls = BigMapType.create_type(args=[key_type, val_type])
        return cls(items=[])  # type: ignore

    @staticmethod
    def from_items(items: List[Tuple[MichelsonType, MichelsonType]]):
        raise AssertionError('forbidden')

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        doc = super(BigMapType, cls).generate_pydoc(definitions, inferred_name=inferred_name)
        return f'{doc} || int /* Big_map ID */'

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'BigMapType':
        return cls(items=[])

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'BigMapType':
        if isinstance(val_expr, dict):
            ptr = parse_micheline_literal(val_expr, {'int': int})
            return cls(items=[], ptr=ptr)
        else:
            items = super(BigMapType, cls).parse_micheline_value(val_expr)
            return cls(items=items)

    @classmethod
    def from_python_object(cls, py_obj: Union[int, dict]) -> 'BigMapType':
        if isinstance(py_obj, int):
            return cls(items=[], ptr=py_obj)
        else:
            items = super(BigMapType, cls).parse_python_object(py_obj)
            return cls(items=items)

    def to_literal(self) -> Type[Micheline]:
        if self.ptr is not None:
            return MichelineLiteral.create(self.ptr)
        else:
            return MichelineSequence.create_type(
                args=[
                    EltLiteral.create_type(
                        args=[
                            k.to_literal(),
                            v.to_literal(),
                        ]
                    )
                    for k, v in self.items
                ]
            )

    def to_micheline_value(self, mode='readable', lazy_diff: Optional[bool] = False):
        if lazy_diff is None:
            lazy_diff = self.ptr is None

        if lazy_diff:
            return super(BigMapType, self).to_micheline_value(mode=mode)
        else:
            assert self.ptr is not None, f'Big_map id is not defined'
            return {'int': str(self.ptr)}

    def to_python_object(self, try_unpack=False, lazy_diff: Optional[bool] = False, comparable=False):
        if lazy_diff is None:
            lazy_diff = self.ptr is None

        if lazy_diff:
            assert not comparable, f'big_map is not comparable'
            res = super(BigMapType, self).to_python_object(try_unpack=try_unpack)
            removals = {key.to_python_object(try_unpack=try_unpack, comparable=True): None for key in self.removed_keys}
            return {**res, **removals}
        else:
            assert self.ptr is not None, f'Big_map id is not defined'
            return self.ptr

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'BigMapType':
        assert self.ptr is not None, f'Big_map id is not defined'
        assert isinstance(lazy_diff, list), f'expected list, got {type(lazy_diff).__name__}'
        diff = next((item for item in lazy_diff if item['kind'] == 'big_map' and item['id'] == str(self.ptr)), None)
        if diff:
            items: List[Tuple[MichelsonType, MichelsonType]] = []
            removed_keys: List[MichelsonType] = []
            for update in diff['diff'].get('updates', []):
                key = self.args[0].from_micheline_value(update['key'])
                if update.get('value'):
                    value = self.args[1].from_micheline_value(update['value'])
                    items.append((key, value))
                else:
                    removed_keys.append(key)
            res = type(self)(ptr=self.ptr, items=items, removed_keys=removed_keys)
            res.context = self.context
            return res
        else:
            return copy(self)

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable') -> 'BigMapType':
        assert self.ptr is not None, f'Big_map ID is not defined'
        if self.context:
            src_ptr, dst_ptr, action = self.context.get_big_map_diff(self.ptr)
        else:
            src_ptr, dst_ptr, action = self.ptr, self.ptr, 'update'

        def make_update(key: MichelsonType, val: Optional[MichelsonType]) -> dict:
            update = {
                'key': key.to_micheline_value(mode=mode),
                'key_hash': forge_script_expr(key.pack(legacy=True)),
            }
            if val is not None:
                update['value'] = val.to_micheline_value(mode=mode)
            return update

        diff = {
            'action': action,
            'updates': [make_update(key, val) for key, val in self],
        }
        if action == 'alloc':
            key_type, val_type = [arg.as_micheline_expr() for arg in self.args]
            diff['key_type'] = key_type  # type: ignore
            diff['value_type'] = val_type  # type: ignore
        elif action == 'copy':
            pass  # TODO:

        lazy_diff.append(
            {
                'kind': 'big_map',
                'id': str(dst_ptr),
                'diff': diff,
            }
        )
        res = type(self)(items=[], ptr=dst_ptr)
        res.context = self.context
        return res

    def find(self, predicate: Callable[['MichelsonType'], bool]) -> Optional['MichelsonType']:
        if predicate(self):
            return self
        return None

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        assert self.context is None, f'context already attached'
        self.context = context
        if self.ptr is None:
            self.ptr = context.get_tmp_big_map_id()
        else:
            self.ptr = context.register_big_map(self.ptr, copy=big_map_copy)
        if context.tzt:  # type: ignore
            context.tzt_big_maps[self.ptr] = self  # type: ignore

    def get(self, key: MichelsonType, dup=True) -> Optional[MichelsonType]:
        self.args[0].assert_type_equal(type(key))
        val = next((v for k, v in self if k == key), Undefined)  # search in diff
        if val is Undefined:
            assert self.context, f'context is not attached'
            key_hash = forge_script_expr(key.pack(legacy=True))
            val_expr = self.context.get_big_map_value(self.ptr, key_hash)  # type: ignore
            if val_expr is None:
                return None
            else:
                return self.args[1].from_micheline_value(val_expr)
        else:
            return val  # type: ignore

    def update(self, key: MichelsonType, val: Optional[MichelsonType]) -> Tuple[Optional[MichelsonType], MichelsonType]:
        removed_keys = set(self.removed_keys)
        prev_val = self.get(key, dup=False)
        if prev_val is not None:
            if val is not None:
                items = [(k, v if k != key else val) for k, v in self]
            else:  # remove
                items = [(k, v) for k, v in self if k != key]
                removed_keys.add(key)
        else:
            if val is not None:
                items = sorted(self.items + [(key, val)], key=lambda x: x[0])
                if key in removed_keys:
                    removed_keys.remove(key)
            else:  # do nothing
                items = self.items  # type: ignore
        res = type(self)(items=items, ptr=self.ptr, removed_keys=list(removed_keys))  # type: ignore
        res.context = self.context
        return prev_val, res

    def get_key_hash(self, key_obj):
        key = self.args[0].from_python_object(key_obj)
        return forge_script_expr(key.pack(legacy=True))

    def duplicate(self):
        res = type(self)(
            items=deepcopy(self.items),
            ptr=self.ptr,
            removed_keys=deepcopy(self.removed_keys),
        )
        res.context = self.context
        return res



from pytezos.rpc.query import RpcQuery


class BakingRightsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/baking_rights'):
    def __call__(self, level=None, cycle=None, delegate=None, max_priority=None, _all=None):
        """Retrieve the list of delegates allowed to bake a block.

        By default, it gives the best baking priorities for bakers that have at least one opportunity
        below the 64th priority for the next block.

        :param level: Specify the (valid) level in the past or future at which the baking rights have to be returned.
        :param cycle: Specify the (valid) levels in the past or future at which the baking rights have to be returned.
        :param delegate: Restrict the results to the given delegates.
        :param max_priority:
        :param _all: If parameter `all` is set, all the baking opportunities for each baker at each level are returned,
            instead of just the first one.
        :returns: Returns the list of baking slots. Also returns the minimal timestamps that correspond to these slots.
            The timestamps are omitted for levels in the past, and are only estimates for levels later that the next block,
            based on the hypothesis that all predecessor blocks were baked at the first priority.
        """
        return self._get(
            params={
                'level': level,
                'cycle': cycle,
                'delegate': delegate,
                'max_priority': max_priority,
                'all': _all,
            }
        )


class ForgeOperationsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/forge/operations'):
    def post(self, operation):
        """Get raw bytes encoding of an unsigned operation

        operation format:

        .. code-block:: python

            {
                "branch": <block hash>,
                "contents": [{ <operation_content> }]
            }

        :param operation: JSON input
        :returns: Hex string
        """
        return self._post(json=operation)


class ForgeProtocolDataQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/forge/protocol_data'):
    def post(self, protocol_data):
        """Forge the protocol-specific part of a block header.

        protocol_data format:

        .. code-block:: python

            {
                "priority": <integer>,
                "nonce_hash": <Base58 nonce hash>,  // optional
                "proof_of_work_nonce": <hex encoded>  // optional
            }

        :param protocol_data: JSON input
        :returns: { "protocol_data": <hex encoded> }
        """
        return self._post(json=protocol_data)


class ForgeBlockHeaderQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/forge_block_header'):
    def post(self, block_header):
        """Forge block header.

        block_header format:

        .. code-block:: python

            {
                "level": <integer>,
                "proto": <integer>,
                "predecessor": <block hash>,
                "timestamp": <timestamp>,
                "validation_pass": <integer>,
                "operations_hash": <Operation_list_list_hash>,
                "fitness": <integer (hex)>,
                "context": <Context_hash>,
                "protocol_data": <hex encoded>
            }


        :param block_header: JSON input
        :returns: { "block": <hex encoded> }
        """
        return self._post(json=block_header)


class ParseBlockQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/parse/block'):
    def post(self, block_header):
        """Retrieve protocol-specific part of a block header and signature.

        block_header format:

        .. code-block:: python

            {
                "level": <integer>,
                "proto": <integer>,
                "predecessor": <block hash>,
                "timestamp": <timestamp>,
                "validation_pass": <integer>,
                "operations_hash": <Operation_list_list_hash>,
                "fitness": <integer (hex)>,
                "context": <Context_hash>,
                "protocol_data": <hex encoded>
            }

        :param block_header: JSON input
        :returns: JSON object
        """
        return self._post(json=block_header)


class ParseOperationsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/parse/operations'):
    def post(self, operations):
        """Extract contents and signatures from the forged and optionally signed operations (bulk).

        operations format:

        .. code-block:: python

            {
                "operations": [ {
                    "branch": <block_hash>,
                    "data": <hex encoded>
                } ... ],
                "check_signature": <boolean>  // optional
            }

        :param operations: JSON input
        :returns: JSON object
        """
        return self._post(json=operations)


class PreapplyBlockQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/preapply/block'):
    def post(self, block, sort=None, timestamp=None):
        """Simulate the validation of a block that would contain the given operations
        and return the resulting fitness and context hash.

        block format:

        .. code-block:: python

            {
                "protocol_data": {
                    "protocol": "Pt24m4xiPbLDhVgVfABUjirbmda3yohdN82Sp9FeuAXJ4eV9otd",
                    "priority": <integer>,
                    "proof_of_work_nonce": <hex encoded>
                    "seed_nonce_hash": <Base58 encoded>,  // optional
                    "signature": <Base58 encoded>
                },
                "operations": [ [ {
                    "protocol": "Pt24m4xiPbLDhVgVfABUjirbmda3yohdN82Sp9FeuAXJ4eV9otd",
                    "branch": <block_hash>,
                    "contents": [ {} ... ],  // kind-specific
                    "signature": <Base58 encoded>
                }... ] ... ]
            }

        :param block: JSON input
        :param sort: by what?
        :param timestamp: timestamp
        :returns: JSON object
        """
        return self._post(
            json=block,
            params={
                'sort': sort,
                'timestamp': timestamp,
            },
        )


class PreapplyOperationsQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/preapply/operations'):
    def post(self, operations):
        """Simulate the validation of operation(s).

        operations format:

        .. code-block:: python

            [{
                "protocol": "Pt24m4xiPbLDhVgVfABUjirbmda3yohdN82Sp9FeuAXJ4eV9otd",
                "branch": <block_hash>,
                "contents": [ {} ... ],  // kind-specific
                "signature": <Base58 encoded>
            }]

        :param operations: JSON input
        :returns: JSON array (operations with metadata)
        """
        return self._post(json=operations)


class ScriptsEntrypoint(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/entrypoint'):
    def post(self, script, entrypoint=None):
        """Return the type of the given entrypoint.

        :param script: Micheline expression
        :param entrypoint: Name of the entrypoint (leave None if there's only one)
        """
        payload = {'script': script}
        if entrypoint:
            payload['entrypoint'] = entrypoint

        return self._post(payload)


class ScriptsEntrypoints(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/entrypoints'):
    def post(self, script):
        """Return the list of entrypoints of the given script.

        :param script: Micheline expression
        """
        return self._post({'script': script})


class ScriptsPackDataQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/pack_data'):
    def post(self, expression):
        """Compute the serialized version of some data expression using the same algorithm as script instruction PACK.

        expression format:

        .. code-block:: python

            {
                "data": <michelson expression>,
                "type": <michelson expression>,
                "gas": <bignum>  // optional
            }

        :param expression: JSON input
        :returns: Packed data (hex encoded)
        """
        return self._post(json=expression)


class ScriptsRunCodeQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/run_code'):
    def post(self, invocation):
        """Run a piece of code in the current context.

        invocation format:

        .. code-block:: python

            {
                "script": <michelson expression>,
                "storage": <michelson expression>,
                "entrypoint": <string>,
                "input": <michelson expression>,
                "chain_id": <base58 encoded>,
                "amount": <mutez>,
                "source": <account address>,  // optional
                "payer": <account address>,  // optional
                "gas": <bignum>  // optional
            }

        :param invocation: JSON input
        :returns: Resulting storage, spawned operations, and bigmap diff
        """
        return self._post(json=invocation)


class ScriptsRunOperationQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/run_operation'):
    def post(self, operation):
        """Run an operation without signature checks.

        operation format:

        .. code-block:: python

            {
                "branch": <block_hash>,
                "contents": [ {} ... ],  // kind-specific
                "signature": <Base58 encoded>
            }

        :param operation: JSON input
        :returns: JSON object (operation with metadata)
        """
        return self._post(json=operation)


class ScriptsTraceCodeQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/trace_code'):
    def post(self, invocation):
        """Run a piece of code in the current context, keeping a trace.

        invocation format:

        .. code-block:: python

            {
                "script": <michelson expression>,
                "storage": <michelson expression>,
                "entrypoint": <string>,
                "input": <michelson expression>,
                "chain_id": <base58 encoded>,
                "amount": <mutez>,
                "source": <account address>,  // optional
                "payer": <account address>,  // optional
                "gas": <bignum>  // optional
            }

        :param invocation: JSON input
        :returns: Resulting storage, spawned operations, bigmap diff, and trace
        """
        return self._post(json=invocation)


class ScriptsTypecheckCodeQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/typecheck_code'):
    def post(self, expression):
        """Check that some data expression is well formed and of a given type in the current context.

        expression format:

        .. code-block:: python

            {
                "program": <michelson expression>,
                "gas": <bignum>  // optional
            }

        :param expression: JSON input
        :returns: JSON object
        """
        return self._post(json=expression)


class ScriptsTypecheckDataQuery(RpcQuery, path='/chains/{}/blocks/{}/helpers/scripts/typecheck_data'):
    def post(self, expression):
        """Check that some data expression is well formed and of a given type in the current context.

        expression format:

        .. code-block:: python

            {
                "data": <michelson expression>,
                "type": <michelson expression>,
                "gas": <bignum>  // optional
            }

        :param expression: JSON input
        :returns: JSON object
        """
        return self._post(json=expression)
